#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import inspect
import logging
from collections import Counter
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd
from pytoda.smiles.transforms import Augment
from pytoda.transforms import AugmentByReversing
from sklearn.utils import shuffle
from terminator.selfies import encoder
from terminator.tokenization import ExpressionBertTokenizer
from transformers.hf_argparser import string_to_bool
from transformers.training_args import TrainingArguments

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

TRANSFORM_FACTORY = {"SELFIES": encoder}
AUGMENT_FACTORY = {
    "SMILES": Augment(),
    "SELFIES": Augment(),
    "AAS": AugmentByReversing(),
}


class Property:
    name: str
    minimum: float = 0
    maximum: float = 0
    expression_separator: str = "|"
    normalize: bool = False

    def __init__(self, name: str):
        self.name = name
        self.mask_lengths: List = []

    def update(self, line: str):
        prop = line.split(self.name)[-1].split(self.expression_separator)[0]
        try:
            val = float(prop)
        except ValueError:
            logger.error(f"Could not convert property {prop} in {line} to float.")
        if val < self.minimum:
            self.minimum = val
        elif val > self.maximum:
            self.maximum = val
        self.mask_lengths.append(len(prop))

    @property
    def mask_length(self) -> int:
        """
        How many tokens are being masked for this property.
        """
        counts = Counter(self.mask_lengths)
        if len(counts) > 1:
            logger.warning(
                f"Not all {self.name} properties have same number of tokens: {counts}"
            )
        return int(counts.most_common(1)[0][0])


def add_tokens_from_lists(
    tokenizer: ExpressionBertTokenizer, train_data: List[str], test_data: List[str]
) -> Tuple[ExpressionBertTokenizer, Dict[str, Property], List[str], List[str]]:
    """
    Addding tokens to a tokenizer from parsed datasets hold in memory.

    Args:
        tokenizer: The tokenizer.
        train_data: List of strings, one per sample.
        test_data: List of strings, one per sample.

    Returns:
       Tuple with:
            tokenizer with updated vocabulary.
            dictionary of property names and full property objects.
            list of strings with training samples.
            list of strings with testing samples.
    """
    num_tokens = len(tokenizer)
    properties: Dict[str, Property] = {}
    all_tokens: Set = set()
    for data in [train_data, test_data]:
        for line in data:
            # Grow the set of all tokens in the dataset
            toks = tokenizer.tokenize(line)
            all_tokens = all_tokens.union(toks)
            # Grow the set of all properties (assumes that the text follows the last `|`)
            props = [
                x.split(">")[0] + ">"
                for x in line.split(tokenizer.expression_separator)[:-1]
            ]
            for prop in props:
                if prop not in properties.keys():
                    properties[prop] = Property(prop)
                properties[prop].update(line)

    # Finish adding new tokens
    tokenizer.add_tokens(list(all_tokens))
    tokenizer.update_vocab(all_tokens)  # type:ignore
    logger.info(f"Added {len(tokenizer)-num_tokens} new tokens to tokenizer.")

    return tokenizer, properties, train_data, test_data


def prepare_datasets_from_files(
    tokenizer: ExpressionBertTokenizer,
    train_path: str,
    test_path: str,
    augment: int = 0,
) -> Tuple[ExpressionBertTokenizer, Dict[str, Property], List[str], List[str]]:
    """
    Converts datasets saved in provided `.csv` paths into RT-compatible datasets.
    NOTE: Also adds the new tokens from train/test data to provided tokenizer.

    Args:
        tokenizer: The tokenizer.
        train_path: Path to the training data.
        test_path: Path to the testing data.
        augment: Factor by which each training sample is augmented.

    Returns:
       Tuple with:
            tokenizer with updated vocabulary.
            dict of property names and property objects.
            list of strings with training samples.
            list of strings with testing samples.
    """

    # Setup data transforms and augmentations
    train_data: List[str] = []
    test_data: List[str] = []
    properties: List[str] = []

    aug = AUGMENT_FACTORY.get(tokenizer.language, lambda x: x)
    trans = TRANSFORM_FACTORY.get(tokenizer.language, lambda x: x)

    for i, (data, path) in enumerate(
        zip([train_data, test_data], [train_path, test_path])
    ):

        if not path.endswith(".csv"):
            raise TypeError(f"Please provide a csv file not {path}.")

        # Load data
        df = shuffle(pd.read_csv(path))
        if "text" not in df.columns:
            raise ValueError("Please provide text in the `text` column.")

        if i == 1 and set(df.columns) != set(properties + ["text"]):
            raise ValueError(
                "Train and test data have to have identical columns, not "
                f"{set(properties + ['text'])} and {set(df.columns)}."
            )
        properties = sorted(list(set(properties).union(list(df.columns))))
        properties.remove("text")

        # Parse data and create RT-compatible format
        for j, row in df.iterrows():
            line = "".join(
                [
                    f"<{p}>{row[p]:.3f}{tokenizer.expression_separator}"
                    for p in properties
                ]
                + [trans(row.text)]  # type: ignore
            )
            data.append(line)

        # Perform augmentation on training data if applicable
        if i == 0 and augment is not None and augment > 1:
            for _ in range(augment):
                for j, row in df.iterrows():
                    line = "".join(
                        [
                            f"<{p}>{row[p]:.3f}{tokenizer.expression_separator}"
                            for p in properties
                        ]
                        + [trans(aug(row.text))]  # type: ignore
                    )
                    data.append(line)

    return add_tokens_from_lists(
        tokenizer=tokenizer, train_data=train_data, test_data=test_data
    )


def get_train_config_dict(
    training_args: Dict[str, Any], properties: Set
) -> Dict[str, Any]:
    return {
        "alternate_steps": training_args["alternate_steps"],
        "reset_training_loss": True,
        "cg_collator": training_args["cg_collator"],
        "cc_loss": training_args["cc_loss"],
        "property_tokens": list(properties),
        "cg_collator_params": {
            "do_sample": False,
            "property_tokens": list(properties),
            "plm_probability": training_args["plm_probability"],
            "max_span_length": training_args["max_span_length"],
            "entity_separator_token": training_args["entity_separator_token"],
            "mask_entity_separator": training_args["mask_entity_separator"],
            "entity_to_mask": training_args["entity_to_mask"],
        },
    }


@dataclass
class TransformersTrainingArgumentsCLI(TrainingArguments):
    """
    GT4SD ships with a CLI to launch training. This conflicts with some data types
    native in `transformers.training_arguments.TrainingArguments` especially iterables
    which cannot be easily passed from CLI.
    Therefore, this class changes the affected attributes to CLI compatible datatypes.
    """

    label_names: Optional[str] = field(  # type: ignore
        default=None,
        metadata={
            "help": "A string containing keys in your dictionary of inputs that correspond to the labels."
            "A single string, but can contain multiple keys separated with comma: `key1,key2`"
        },
    )
    report_to: Optional[str] = field(  # type: ignore
        default=None,
        metadata={
            "help": "The list of integrations to report the results and logs to."
            "A single string, but can contain multiple keys separated with comma: `i1,i2`"
        },
    )
    sharded_ddp: str = field(
        default="",
        metadata={
            "help": "Whether or not to use sharded DDP training (in distributed training only). The base option "
            "should be `simple`, `zero_dp_2` or `zero_dp_3` and you can add CPU-offload to `zero_dp_2` or `zero_dp_3` "
            "like this: zero_dp_2 offload` or `zero_dp_3 offload`. You can add auto-wrap to `zero_dp_2` or "
            "with the same syntax: zero_dp_2 auto_wrap` or `zero_dp_3 auto_wrap`.",
        },
    )
    tf32: Optional[str] = field(  # type: ignore
        default="no",
        metadata={
            "help": (
                "Whether to enable tf32 mode, available in Ampere and newer GPU architectures. This is an experimental"
                " API and it may change."
            )
        },
    )
    disable_tqdm: Optional[str] = field(  # type: ignore
        default="no",
        metadata={"help": "Whether or not to disable the tqdm progress bars."},
    )
    greater_is_better: Optional[str] = field(  # type: ignore
        default="no",
        metadata={
            "help": "Whether the `metric_for_best_model` should be maximized or not."
        },
    )
    remove_unused_columns: Optional[str] = field(  # type: ignore
        default="yes",
        metadata={
            "help": "Remove columns not required by the model when using an nlp.Dataset."
        },
    )
    load_best_model_at_end: Optional[str] = field(  # type: ignore
        default=None,
        metadata={
            "help": "Whether or not to load the best model found during training at the end of training."
        },
    )
    ddp_find_unused_parameters: Optional[str] = field(  # type: ignore
        default="no",
        metadata={
            "help": (
                "When using distributed training, the value of the flag `find_unused_parameters` passed to "
                "`DistributedDataParallel`."
            )
        },
    )
    evaluation_strategy: Optional[str] = field(  # type: ignore
        default="no",
        metadata={
            "help": (
                "The evaluation strategy to adopt during training. Possible values are:"
                " - 'no': No evaluation is done during training."
                " - 'steps'`: Evaluation is done (and logged) every `eval_steps`."
                " - 'epoch': Evaluation is done at the end of each epoch."
            )
        },
    )
    lr_scheduler_type: Optional[str] = field(  # type: ignore
        default="linear",
        metadata={
            "help": (
                "The scheduler type to use. See the documentation of "
                "`transformers.SchedulerType` for all possible values."
            )
        },
    )
    logging_strategy: Optional[str] = field(  # type: ignore
        default="steps",
        metadata={
            "help": (
                "The logging strategy to adopt during training. Possible values are:"
                " - 'no': No logging is done during training."
                " - 'steps'`: Logging is done every `logging_steps`."
                " - 'epoch': Logging is done at the end of each epoch."
            )
        },
    )
    save_strategy: Optional[str] = field(  # type: ignore
        default="steps",
        metadata={
            "help": (
                "The saving strategy to adopt during training. Possible values are:"
                " - 'no': No saving is done during training."
                " - 'steps'`: Saving is done every `saving_steps`."
                " - 'epoch': Saving is done at the end of each epoch."
            )
        },
    )
    hub_strategy: Optional[str] = field(  # type: ignore
        default="every_save",
        metadata={
            "help": (
                "Optional, defaults to `every_save`. Defines the scope of what is pushed "
                "to the Hub and when. Possible values are:"
                " - `end`: push the model, its configuration, the tokenizer (if passed "
                "       along to the Trainer and a draft of a model card when the "
                "       Trainer.save_model method is called."
                " - `every_save`: push the model, its configuration, the tokenizer (if "
                "       passed along to the Trainer and a draft of a model card each time "
                "       there is a model save. The pushes are asynchronous to not block "
                "       training, and in case the save are very frequent, a new push is "
                "       only attempted if the previous one is finished. A last push is made "
                "       with the final model at the end of training."
                " - `checkpoint`: like `every_save` but the latest checkpoint is also "
                "       pushed in a subfolder named last-checkpoint, allowing you to resume "
                "       training easily with `trainer.train(resume_from_checkpoint)`."
                " - `all_checkpoints`: like `checkpoint` but all checkpoints are pushed "
                "       like they appear in the output folder (so you will get one "
                "       checkpoint folder per folder in your final repository)."
            )
        },
    )
    optim: Optional[str] = field(  # type: ignore
        default="adamw_hf",
        metadata={
            "help": (
                "The optimizer to use. One of  `adamw_hf`, `adamw_torch`, `adafactor` "
                "or `adamw_apex_fused`."
            )
        },
    )

    def __post_init__(self):
        """
        Necessary because the our ArgumentParser (that is based on argparse) converts
        empty strings to None. This is prohibitive since the HFTrainer relies on
        them being actual strings. Only concerns a few arguments.
        """
        if self.sharded_ddp is None:
            self.sharded_ddp = ""
        if self.fsdp is None:  # type: ignore
            self.fsdp = ""

        self.disable_tqdm = string_to_bool(self.disable_tqdm)
        self.tf32 = string_to_bool(self.tf32)

        super().__post_init__()


def get_hf_training_arg_object(training_args: Dict[str, Any]) -> TrainingArguments:
    """
    A method to convert a training_args Dictionary into a HuggingFace
    `TrainingArguments` object.
    This routine also takes care of removing arguments that are not necessary.

    Args:
        training_args: A dictionary of training arguments.

    Returns:
        object of type `TrainingArguments`.
    """

    # Get attributes of parent class
    org_attrs = dict(inspect.getmembers(TrainingArguments))

    # Remove attributes that were specified by child classes
    hf_training_args = {k: v for k, v in training_args.items() if k in org_attrs.keys()}

    # Instantiate class object
    hf_train_object = TrainingArguments(training_args["output_dir"])

    # Set attributes manually (since this is a `dataclass` not everything can be passed
    # to constructor)
    for k, v in hf_training_args.items():
        setattr(hf_train_object, k, v)

    return hf_train_object



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Regression Transformer training utilities."""
import os
from dataclasses import dataclass, field
from typing import Optional

from ...configuration import gt4sd_configuration_instance
from ..core import TrainingPipelineArguments
from .utils import TransformersTrainingArgumentsCLI

DATA_ROOT = os.path.join(
    gt4sd_configuration_instance.gt4sd_local_cache_path, "data", "RegressionTransformer"
)
os.makedirs(DATA_ROOT, exist_ok=True)


@dataclass
class RegressionTransformerTrainingArguments(
    TrainingPipelineArguments, TransformersTrainingArgumentsCLI
):
    """
    Arguments related to RegressionTransformer trainer.
    NOTE: All arguments from `transformers.training_args.TrainingArguments` can be used.
    Only additional ones are specified below.
    """

    __name__ = "training_args"

    training_name: str = field(
        default="rt_training", metadata={"help": "Name used to identify the training."}
    )
    num_train_epochs: int = field(default=10, metadata={"help": "Number of epochs."})
    batch_size: int = field(default=16, metadata={"help": "Size of the batch."})
    log_interval: int = field(
        default=100, metadata={"help": "Number of steps between log intervals."}
    )
    gradient_interval: int = field(
        default=1, metadata={"help": "Gradient accumulation steps"}
    )
    eval_steps: int = field(
        default=1000,
        metadata={"help": "The time interval at which validation is performed."},
    )

    max_span_length: int = field(
        default=5, metadata={"help": "Max length of a span of masked tokens for PLM."}
    )
    plm_probability: float = field(
        default=1 / 6,
        metadata={
            "help": "Ratio of length of a span of masked tokens to surrounding context length for PLM."
        },
    )
    alternate_steps: int = field(
        default=50,
        metadata={
            "help": "Per default, training alternates between property prediction and "
            "conditional generation. This argument specifies the alternation frequency."
            "If you set it to 0, no alternation occurs and we fall back to vanilla "
            "permutation language modeling (PLM). Default: 50."
        },
    )
    cc_loss: bool = field(
        default=True,
        metadata={
            "help": "Whether the cycle-consistency loss is computed during the conditional "
            "generation task. Defaults to True."
        },
    )
    cg_collator: str = field(
        default="vanilla_cg",
        metadata={
            "help": "The collator class. Following options are implemented: "
            "'vanilla_cg': Collator class that does not mask the properties but anything else as a regular DataCollatorForPermutationLanguageModeling. Can optionally replace the properties with sampled values. "
            "NOTE: This collator can deal with multiple properties. "
            "'multientity_cg': A training collator the conditional-generation task that can handle multiple entities. "
            "Default: vanilla_cg."
        },
    )
    entity_to_mask: int = field(
        default=-1,
        metadata={
            "help": "Only applies if `cg_collator='multientity_cg'`. The entity that is being masked during training. 0 corresponds to first entity and so on. -1 corresponds to "
            "a random sampling scheme where the entity-to-be-masked is determined "
            "at runtime in the collator. NOTE: If 'mask_entity_separator' is true, "
            "this argument will not have any effect. Defaults to -1."
        },
    )
    entity_separator_token: str = field(
        default=".",
        metadata={
            "help": "Only applies if `cg_collator='multientity_cg'`.The token that is used to separate "
            "entities in the input. Defaults to '.' (applicable to SMILES & SELFIES)"
        },
    )
    mask_entity_separator: bool = field(
        default=False,
        metadata={
            "help": "Only applies if `cg_collator='multientity_cg'`. Whether or not the entity separator token can be masked. If True, *all** textual tokens can be masked and we "
            "the collator behaves like the `vanilla_cg ` even though it is a `multientity_cg`. If False, the exact behavior "
            "depends on the entity_to_mask argument. Defaults to False."
        },
    )


@dataclass
class RegressionTransformerDataArguments(TrainingPipelineArguments):
    """Arguments related to RegressionTransformer data loading."""

    __name__ = "dataset_args"

    train_data_path: str = field(
        metadata={
            "help": "Path to a `.csv` file with the input training data. The file has to "
            "contain a `text` column (with the string input, e.g, SMILES, AAS, natural "
            "text) and an arbitrary number of numerical columns."
        },
    )
    test_data_path: str = field(
        metadata={
            "help": "Path to a `.csv` file with the input testing data. The file has to "
            "contain a `text` column (with the string input, e.g, SMILES, AAS, natural "
            "text) and an arbitrary number of numerical columns."
        },
    )
    augment: Optional[int] = field(
        default=0,
        metadata={
            "help": "Factor by which the training data is augmented. The data modality "
            "(SMILES, SELFIES, AAS, natural text) is inferred from the tokenizer. "
            "NOTE: For natural text, no augmentation is supported. Defaults to 0, "
            "meaning no augmentation. "
        },
    )
    save_datasets: Optional[bool] = field(
        default=False,
        metadata={
            "help": "Whether to save the datasets to disk. Datasets will be saved as `.txt` file to "
            "the same location where `train_data_path` and `test_data_path` live. Defaults to False."
        },
    )


@dataclass
class RegressionTransformerSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to RegressionTransformer trainer."""

    __name__ = "saving_args"

    model_path: str = field(
        metadata={"help": "Path where the model artifacts are stored."}
    )
    checkpoint_name: str = field(
        default=str(),
        metadata={
            "help": "Name for the checkpoint that should be copied to inference model. "
            "Has to be a subfolder of `model_path`. Defaults to empty string meaning that "
            "files are taken from `model_path` (i.e., after training finished)."
        },
    )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Regression Transformer trainer unit tests."""

import json
import os
import tempfile
from typing import Any, Dict, Iterable, cast

import importlib_resources

from gt4sd.algorithms.conditional_generation.regression_transformer import (
    RegressionTransformerMolecules,
)
from gt4sd.cli.argument_parser import DataClassType
from gt4sd.cli.trainer import TrainerArgumentParser
from gt4sd.training_pipelines import (
    TRAINING_PIPELINE_ARGUMENTS_MAPPING,
    TRAINING_PIPELINE_MAPPING,
    RegressionTransformerTrainingPipeline,
)
from gt4sd.training_pipelines.core import TrainingPipelineArguments

template_config = {
    "model_args": {},
    "dataset_args": {},
    "training_args": {
        "training_name": "regression-transformer-test",
        "batch_size": 4,
        "learning_rate": 0.0005,
        "do_train": True,
        "eval_accumulation_steps": 1,
        "eval_steps": 100,
        "save_steps": 100,
        "max_steps": 3,
        "epochs": 1,
        "overwrite_output_dir": True,
    },
}

xlnet_config = {
    "architectures": ["XLNetLMHeadModel"],
    "attn_type": "bi",
    "bi_data": False,
    "bos_token_id": 14,
    "clamp_len": -1,
    "d_head": 8,
    "d_inner": 512,
    "d_model": 32,
    "dropout": 0.2,
    "end_n_top": 5,
    "eos_token_id": 14,
    "ff_activation": "gelu",
    "initializer_range": 0.02,
    "language": "selfies",
    "layer_norm_eps": 1e-12,
    "model_type": "xlnet",
    "n_head": 4,
    "n_layer": 8,
    "ne_dim": 16,
    "ne_format": "sum",
    "ne_type": "float",
    "pad_token_id": 0,
    "same_length": False,
    "start_n_top": 5,
    "summary_activation": "tanh",
    "summary_last_dropout": 0.1,
    "summary_type": "last",
    "summary_use_proj": True,
    "task_specific_params": {"text-generation": {"do_sample": True, "max_length": 250}},
    "untie_r": True,
    "use_ne": True,
    "vmax": 1.0,
}


def combine_defaults_and_user_args(
    config: Dict[str, Dict[str, Any]]
) -> Dict[str, Dict[str, Any]]:

    arguments = TRAINING_PIPELINE_ARGUMENTS_MAPPING["regression-transformer-trainer"]
    """
    We need `conflict_handler='resolve'` because the RT relies on the TrainingArguments
    in HuggingFace. Some arguments, like `output_dir` do not have defaults and thus
    an "empty" parser like the below complains once we call `parse_args_into_dataclasses`.
    Therefore, we manually add the `output_dir` argument with the correct value.
    """
    parser = TrainerArgumentParser(
        cast(
            Iterable[DataClassType],
            tuple([*arguments]),
        ),
        conflict_handler="resolve",
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        help="The output directory where the model predictions and checkpoints will be written.",
        default=config["training_args"]["output_dir"],
    )
    parser.add_argument(
        "--train_data_path",
        type=str,
        help="The output directory where the model predictions and checkpoints will be written.",
        default=config["dataset_args"]["train_data_path"],
    )
    parser.add_argument(
        "--test_data_path",
        type=str,
        help="The output directory where the model predictions and checkpoints will be written.",
        default=config["dataset_args"]["test_data_path"],
    )
    args = parser.parse_args_into_dataclasses(return_remaining_strings=True)
    input_config = {
        arg.__name__: arg.__dict__
        for arg in args
        if isinstance(arg, TrainingPipelineArguments) and isinstance(arg.__name__, str)
    }
    input_config["model_args"].update(**config["model_args"])
    input_config["dataset_args"].update(**config["dataset_args"])
    input_config["training_args"].update(**config["training_args"])

    return input_config


def test_train():

    pipeline = TRAINING_PIPELINE_MAPPING.get("regression-transformer-trainer")
    assert pipeline is not None

    mol_model = RegressionTransformerMolecules(algorithm_version="qed")
    mol_path = mol_model.ensure_artifacts_for_version("qed")

    TEMPORARY_DIRECTORY = tempfile.mkdtemp()
    test_pipeline = cast(RegressionTransformerTrainingPipeline, pipeline())

    config: Dict[str, Any] = template_config.copy()
    config["training_args"]["output_dir"] = TEMPORARY_DIRECTORY
    with importlib_resources.as_file(
        importlib_resources.files("gt4sd")
        / "training_pipelines/tests/regression_transformer_raw.csv"
    ) as raw_path:

        # Test finetuning the QED model
        config["model_args"]["model_path"] = mol_path
        config["dataset_args"]["train_data_path"] = str(raw_path)
        config["dataset_args"]["test_data_path"] = str(raw_path)
        config["dataset_args"]["augment"] = 2
        input_config = combine_defaults_and_user_args(config)
        test_pipeline.train(**input_config)

        # Test training model from scratch
        with tempfile.TemporaryDirectory() as temp:
            f_name = os.path.join(temp, "tmp_xlnet_config.json")
            # Write file
            with open(f_name, "w") as f:
                json.dump(xlnet_config, f, indent=4)

            config["model_args"]["config_name"] = f_name
            del config["model_args"]["model_path"]
            config["model_args"]["tokenizer_name"] = mol_path
            config["dataset_args"]["data_path"] = str(raw_path)
            config["dataset_args"]["augment"] = 2
            input_config = combine_defaults_and_user_args(config)
            test_pipeline.train(**input_config)

            # Test the multientity collator (finetuning QED model)
            config["training_args"].update(
                {
                    "cg_collator": "multientity_cg",
                    "entity_to_mask": 0,
                    "entity_separator_token": ".",
                }
            )
            input_config = combine_defaults_and_user_args(config)
            test_pipeline.train(**input_config)

    with importlib_resources.as_file(
        importlib_resources.files("gt4sd")
        / "training_pipelines/tests/regression_transformer_copolymer_raw.csv"
    ) as raw_path:

        # Test finetuning the polymer model
        polymer_model = RegressionTransformerMolecules(
            algorithm_version="block_copolymer"
        )
        polymer_path = polymer_model.ensure_artifacts_for_version("block_copolymer")

        config["model_args"]["model_path"] = polymer_path
        config["dataset_args"]["train_data_path"] = str(raw_path)
        config["dataset_args"]["test_data_path"] = str(raw_path)
        input_config = combine_defaults_and_user_args(config)
        test_pipeline.train(**input_config)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import Dict, Tuple, Type

from ..core import PropertyPredictor, PropertyPredictorParameters
from .core import (
    AliphaticIndex,
    AmideConfiguration,
    AmidePhConfiguration,
    Aromaticity,
    BomanIndex,
    Charge,
    ChargeDensity,
    HydrophobicRatio,
    Instability,
    IsoelectricPoint,
    Length,
    MolecularWeight,
)

# NOTE: all functions can be called with either an AA sequence or a rdkit.Chem.Mol object.
PROTEIN_PROPERTY_PREDICTOR_FACTORY: Dict[
    str, Tuple[Type[PropertyPredictor], Type[PropertyPredictorParameters]]
] = {
    # inherent properties
    "length": (Length, PropertyPredictorParameters),
    "protein_weight": (MolecularWeight, AmideConfiguration),
    # rule-based properties
    "boman_index": (BomanIndex, PropertyPredictorParameters),
    "charge_density": (ChargeDensity, AmidePhConfiguration),
    "charge": (Charge, AmidePhConfiguration),
    "aliphaticity": (AliphaticIndex, PropertyPredictorParameters),
    "hydrophobicity": (HydrophobicRatio, PropertyPredictorParameters),
    "isoelectric_point": (IsoelectricPoint, AmideConfiguration),
    "aromaticity": (Aromaticity, PropertyPredictorParameters),
    "instability": (Instability, PropertyPredictorParameters),
    # properties predicted by ML models
}

AVAILABLE_PROTEINS_PROPERTY_PREDICTOR = sorted(
    PROTEIN_PROPERTY_PREDICTOR_FACTORY.keys()
)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy
import enum
from typing import Any, Dict, List, Tuple, Union

import networkx as nx
import numpy as np
import torch
import torch_geometric.data as gd
from rdkit.Chem import Mol
from torch_scatter import scatter, scatter_max


class Graph(nx.Graph):
    """Subclassing nx.Graph for debugging purposes."""

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f'<{list(self.nodes)}, {list(self.edges)}, {list(self.nodes[i]["v"] for i in self.nodes)}>'


def graph_without_edge(g: nx.Graph, e: nx.Graph) -> nx.Graph:
    """Build a new graph without an edge.

    Args:
        g: a graph.
        e: an edge to remove.

    Returns:
        a new graph without an edge.
    """

    gp = g.copy()
    gp.remove_edge(*e)
    return gp


def graph_without_node(g: nx.Graph, n: nx.Graph) -> nx.Graph:
    """Build a new graph without a node.

    Args:
        g: a graph.
        n: a node to remove.

    Returns:
        a new graph without a node.
    """

    gp = g.copy()
    gp.remove_node(n)
    return gp


def graph_without_node_attr(g: nx.Graph, n: nx.Graph, a: Any) -> nx.Graph:
    """Build a new graph without a node attribute.

    Args:
        g: a graph.
        n: a node.
        a: a node attribute.

    Returns:
        a new graph without a node attribute.
    """

    gp = g.copy()
    del gp.nodes[n][a]
    return gp


def graph_without_edge_attr(g: nx.Graph, e: nx.Graph, a: Any) -> nx.Graph:
    """Build a new graph without an edge attribute.

    Args:
        g: a graph.
        e: an edge.
        a: an edge attribute.

    Returns:
        a new graph without an edge attribute.
    """

    gp = g.copy()
    del gp.edges[e][a]
    return gp


class GraphActionType(enum.Enum):
    """Action types for graph building environments."""

    # Forward actions
    Stop = enum.auto()
    AddNode = enum.auto()
    AddEdge = enum.auto()
    SetNodeAttr = enum.auto()
    SetEdgeAttr = enum.auto()
    # Backward actions
    RemoveNode = enum.auto()
    RemoveEdge = enum.auto()
    RemoveNodeAttr = enum.auto()
    RemoveEdgeAttr = enum.auto()


class GraphAction:
    """Actions on a graph environment for gflownet."""

    def __init__(
        self,
        action: GraphActionType,
        source: int = None,
        target: int = None,
        value: Any = None,
        attr: str = None,
        relabel: int = None,
    ):
        """Initialize a single graph-building action.

        Args:
            action: the action type.
            source: the source node this action is applied on.
            target: the target node (i.e. if specified this is an edge action).
            value: the value (e.g. new node type) applied.
            attr: the set attribute of a node/edge.
            relabel: for AddNode actions, relabels the new node with that id.
        """
        self.action = action
        self.source = source
        self.target = target
        self.attr = attr
        self.value = value
        self.relabel = relabel  # TODO: deprecate this?

    def __repr__(self):
        attrs = ", ".join(
            str(i)
            for i in [self.source, self.target, self.attr, self.value]
            if i is not None
        )
        return f"<{self.action}, {attrs}>"


class GraphBuildingEnv:
    """A Graph building environment which induces a dag state space compatible with gflownet."""

    def __init__(
        self,
        allow_add_edge: bool = True,
        allow_node_attr: bool = True,
        allow_edge_attr: bool = True,
    ):
        """Initialize a GraphBuildingEnv.

        Supports forward and backward actions, with a `parents` function that list parents of
        forward actions. Edges and nodes can have attributes added to them in a key:value style.
        Edges and nodes are created with _implicit_ default attribute
        values (e.g. chirality, single/double bondness) so that:
            - an agent gets to do an extra action to set that attribute, but only
            if it is still default-valued (DAG property preserved)
            - we can generate a legal action for any attribute that isn't a default one.

        Code adapted from: https://github.com/recursionpharma/gflownet/tree/trunk/src/gflownet/envs

        Args:
            allow_add_edge: if True, allows this action and computes AddEdge parents (i.e. if False, this
                env only allows for tree generation).
            allow_node_attr: if True, allows this action and computes SetNodeAttr parents.
            allow_edge_attr: if True, allows this action and computes SetEdgeAttr parents.
        """

        self.allow_add_edge = allow_add_edge
        self.allow_node_attr = allow_node_attr
        self.allow_edge_attr = allow_edge_attr

    def new(self) -> nx.Graph:
        return Graph()

    def step(self, g: Graph, action: GraphAction) -> Graph:
        """Step forward the given graph state with an action

        Args:
            g: the graph to be modified.
            action: the action taken on the graph, indices must match.

        Returns:
            gp: the new graph.
        """
        gp = g.copy()
        if action.action is GraphActionType.AddEdge:
            a, b = action.source, action.target

            assert self.allow_add_edge
            assert a in g and b in g

            if a > b:  # type: ignore
                a, b = b, a

            assert a != b
            assert not g.has_edge(a, b)

            # Ideally the FA underlying this must only be able to send
            # create_edge actions which respect this a<b property (or
            # its inverse!) , otherwise symmetry will be broken
            # because of the way the parents method is written
            gp.add_edge(a, b)

        elif action.action is GraphActionType.AddNode:
            if len(g) == 0:
                assert action.source == 0  # TODO: this may not be useful
                gp.add_node(0, v=action.value)
            else:
                assert action.source in g.nodes
                e = [action.source, max(g.nodes) + 1]
                if action.relabel is not None:
                    raise ValueError("deprecated")
                # if kw and 'relabel' in kw:
                #     e[1] = kw['relabel']  # for `parent` consistency, allow relabeling
                assert not g.has_edge(*e)
                gp.add_node(e[1], v=action.value)
                gp.add_edge(*e)

        elif action.action is GraphActionType.SetNodeAttr:
            assert self.allow_node_attr
            assert action.source in gp.nodes
            assert action.attr not in gp.nodes[action.source]
            gp.nodes[action.source][action.attr] = action.value

        elif action.action is GraphActionType.SetEdgeAttr:
            assert self.allow_edge_attr
            assert g.has_edge(action.source, action.target)
            assert action.attr not in gp.edges[(action.source, action.target)]
            gp.edges[(action.source, action.target)][action.attr] = action.value
        else:
            # TODO: backward actions if we want to support MCMC-GFN style algorithms
            raise ValueError(f"Unknown action type {action.action}", action.action)

        return gp

    def count_backward_transitions(self, g: Graph) -> int:
        """Counts the number of parents of g without checking for isomorphisms."""
        c = 0
        deg = [g.degree[i] for i in range(len(g.nodes))]
        for a, b in g.edges:
            if deg[a] > 1 and deg[b] > 1 and len(g.edges[(a, b)]) == 0:
                # Can only remove edges connected to non-leaves and without
                # attributes (the agent has to remove the attrs, then remove
                # the edge). Removal cannot disconnect the graph.
                new_g = graph_without_edge(g, (a, b))
                if nx.algorithms.is_connected(new_g):
                    c += 1
            c += len(g.edges[(a, b)])  # One action per edge attr
        for i in g.nodes:
            if (
                deg[i] == 1
                and len(g.nodes[i]) == 1
                and len(g.edges[list(g.edges(i))[0]]) == 0
            ):
                c += 1
            c += len(g.nodes[i]) - 1  # One action per node attr, except 'v'
            if len(g.nodes) == 1 and len(g.nodes[i]) == 1:
                # special case if last node in graph
                c += 1
        return c


def generate_forward_trajectory(
    g: Graph, max_nodes: int = None
) -> List[Tuple[Graph, GraphAction]]:
    """Sample (uniformly) a trajectory that generates g.

    Args:
        g: the graph to be generated.
        max_nodes: the maximum number of nodes to generate.

    Returns:
        a list of (graph, action) pairs.
    """

    gn = Graph()
    # Choose an arbitrary starting point, add to the stack
    stack: List[Tuple[int, ...]] = [(np.random.randint(0, len(g.nodes)),)]
    traj = []
    # This map keeps track of node labels in gn, since we have to start from 0
    relabeling_map: Dict[int, int] = {}
    while len(stack):
        # We pop from the stack until all nodes and edges have been
        # generated and their attributes have been set. Uninserted
        # nodes/edges will be added to the stack as the graph is
        # expanded from the starting point. Nodes/edges that have
        # attributes will be reinserted into the stack until those
        # attributes are "set".
        i = stack.pop(np.random.randint(len(stack)))

        gt = gn.copy()  # This is a shallow copy

        if len(i) > 1:  # i is an edge
            e = relabeling_map.get(i[0], None), relabeling_map.get(i[1], None)
            if e in gn.edges:
                # i exists in the new graph, that means some of its attributes need to be added
                attrs = [j for j in g.edges[i] if j not in gn.edges[e]]
                if len(attrs) == 0:
                    continue  # If nodes are in cycles edges leading to them get stack multiple times, disregard
                attr = attrs[np.random.randint(len(attrs))]
                gn.edges[e][attr] = g.edges[i][attr]
                act = GraphAction(
                    GraphActionType.SetEdgeAttr,
                    source=e[0],
                    target=e[1],
                    attr=attr,
                    value=g.edges[i][attr],
                )
            else:
                # i doesn't exist, add the edge
                if e[1] not in gn.nodes:
                    # The endpoint of the edge is not in the graph, this is a AddNode action
                    assert e[1] is None  # normally we shouldn't have relabeled i[1] yet
                    relabeling_map[i[1]] = len(relabeling_map)
                    e = e[0], relabeling_map[i[1]]
                    gn.add_node(e[1], v=g.nodes[i[1]]["v"])
                    gn.add_edge(*e)
                    for j in g[i[1]]:  # stack unadded edges/neighbours
                        jp = relabeling_map.get(j, None)
                        if jp not in gn or (e[1], jp) not in gn.edges:
                            stack.append((i[1], j))
                    act = GraphAction(
                        GraphActionType.AddNode, source=e[0], value=g.nodes[i[1]]["v"]
                    )
                    if len(gn.nodes[e[1]]) < len(g.nodes[i[1]]):
                        stack.append(
                            (i[1],)
                        )  # we still have attributes to add to node i[1]
                else:
                    # The endpoint is in the graph, this is an AddEdge action
                    assert e[0] in gn.nodes
                    gn.add_edge(*e)
                    act = GraphAction(GraphActionType.AddEdge, source=e[0], target=e[1])

            if len(gn.edges[e]) < len(g.edges[i]):
                stack.append(i)  # we still have attributes to add to edge i
        else:  # i is a node, (u,)
            u = i[0]
            n = relabeling_map.get(u, None)
            if n not in gn.nodes:
                # u doesn't exist yet, this should only happen for the first node
                assert len(gn.nodes) == 0
                act = GraphAction(
                    GraphActionType.AddNode, source=0, value=g.nodes[u]["v"]
                )
                n = relabeling_map[u] = len(relabeling_map)
                gn.add_node(0, v=g.nodes[u]["v"])
                for j in g[u]:  # For every neighbour of node u
                    if relabeling_map.get(j, None) not in gn:
                        stack.append((u, j))  # push the (u,j) edge onto the stack
            else:
                # u exists, meaning we have attributes left to add
                attrs = [j for j in g.nodes[u] if j not in gn.nodes[n]]
                attr = attrs[np.random.randint(len(attrs))]
                gn.nodes[n][attr] = g.nodes[u][attr]
                act = GraphAction(
                    GraphActionType.SetNodeAttr,
                    source=n,
                    attr=attr,
                    value=g.nodes[u][attr],
                )
            if len(gn.nodes[n]) < len(g.nodes[u]):
                stack.append((u,))  # we still have attributes to add to node u
        traj.append((gt, act))
    traj.append((gn, GraphAction(GraphActionType.Stop)))
    return traj


class GraphActionCategorical:
    """A multi-type categorical action class compatible with generating structured actions on a graph."""

    def __init__(
        self,
        graphs: gd.Batch,
        logits: List[torch.Tensor],
        keys: List[Union[str, None]],
        types: List[GraphActionType],
        deduplicate_edge_index=True,
    ) -> None:
        """Initialize a GraphActionCategorical.

        What is meant by multi-type here is that there are multiple types of
        mutually exclusive actions, e.g. AddNode and AddEdge are
        mutually exclusive, but since their logits will be produced by
        different variable-sized tensors (corresponding to different
        elements of the graph, e.g. nodes or edges) it is inconvient
        to stack them all into one single Categorical. This class
        provides this convenient interaction between torch_geometric
        Batch objects and lists of logit tensors.

        Code adapted from: https://github.com/recursionpharma/gflownet/tree/trunk/src/gflownet/envs

        Args:
            graphs: a Batch of graphs to which the logits correspond.
            logits: a list of tensors of shape `(n, m)` representing logits
                over a variable number of graph elements (e.g. nodes) for
                which there are `m` possible actions. `n` should thus be
                equal to the sum of the number of such elements for each
                graph in the Batch object. The length of the `logits` list
                should thus be equal to the number of element types (in
                other words there should be one tensor per type).
            keys: the keys corresponding to the Graph elements for each
                tensor in the logits list. Used to extract the `_batch`
                and slice attributes. For example, if the first logit
                tensor is a per-node action logit, and the second is a
                per-edge, `keys` could be `['x', 'edge_index']`. If
                keys[i] is None, the corresponding logits are assumed to
                be graph-level (i.e. if there are `k` graphs in the Batch
                object, this logit tensor would have shape `(k, m)`).
            types: the action type each logit corresponds to.
            deduplicate_edge_index: if true, this means that the 'edge_index' keys have been reduced
                by e_i[::2] (presumably because the graphs are undirected).
        """

        self.num_graphs = graphs.num_graphs
        # The logits
        self.logits = logits
        self.types = types
        self.keys = keys
        self.dev = dev = graphs.x.device

        # This is the minibatch index of each entry in the logits
        # i.e., if graph i in the Batch has N[i] nodes,
        #    g.batch == [0,0,0, ...,  1,1,1,1,1, ... ]
        #                 N[0] times    N[1] times
        # This generalizes to edges and non-edges.
        # Append '_batch' to keys except for 'x', since TG has a special case (done by default for 'x')
        self.batch = [
            getattr(graphs, f"{k}_batch" if k != "x" else "batch") if k is not None
            # None signals a global logit rather than a per-instance logit
            else torch.arange(graphs.num_graphs, device=dev)
            for k in keys
        ]
        # This is the cumulative sum (prefixed by 0) of N[i]s
        self.slice = [
            graphs._slice_dict[k]
            if k is not None
            else torch.arange(graphs.num_graphs, device=dev)
            for k in keys
        ]
        self.logprobs: Union[List[Any], None] = None

        if deduplicate_edge_index and "edge_index" in keys:
            idx = keys.index("edge_index")
            self.batch[idx] = self.batch[idx][::2]
            self.slice[idx] = self.slice[idx].div(2, rounding_mode="floor")

    def detach(self):
        """Detach the logits from the graph batch."""
        new = copy.copy(self)
        new.logits = [i.detach() for i in new.logits]
        if new.logprobs is not None:
            new.logprobs = [i.detach() for i in new.logprobs]
        return new

    def to(self, device):
        """Move everything to the specified device.

        Args:
            device: the device to move to.
        """
        self.dev = device
        self.logits = [i.to(device) for i in self.logits]
        self.batch = [i.to(device) for i in self.batch]
        self.slice = [i.to(device) for i in self.slice]
        if self.logprobs is not None:
            self.logprobs = [i.to(device) for i in self.logprobs]
        return self

    def logsoftmax(self) -> List[torch.Tensor]:
        """Compute log-probabilities given logits.

        Returns:
            a list of log-probabilities.
        """

        if self.logprobs is not None:
            return self.logprobs
        # Use the `subtract by max` trick to avoid precision errors:
        # compute max
        maxl = (
            torch.cat(
                [
                    scatter(i, b, dim=0, dim_size=self.num_graphs, reduce="max")
                    for i, b in zip(self.logits, self.batch)
                ],
                dim=1,
            )
            .max(1)
            .values.detach()
        )

        # substract by max then take exp
        # x[b, None] indexes by the batch to map back to each node/edge and adds a broadcast dim
        exp_logits = [
            (i - maxl[b, None]).exp() + 1e-40 for i, b in zip(self.logits, self.batch)
        ]

        # sum corrected exponentiated logits, to get log(z - max) = log(sum(exp(logits)) - max)
        log_z = sum(  # type: ignore
            [
                scatter(i, b, dim=0, dim_size=self.num_graphs, reduce="sum").sum(1)
                for i, b in zip(exp_logits, self.batch)
            ]
        ).log()

        # log probabilities is log(exp(logit) / Z)
        self.logprobs = [
            i.log() - log_z[b, None] for i, b in zip(exp_logits, self.batch)
        ]
        return self.logprobs

    def sample(self) -> List[Tuple[int, int, int]]:
        """Use the Gumbel trick to sample categoricals.

        if X ~ argmax(logits - log(-log(uniform(logits.shape))))
        then  p(X = i) = exp(logits[i]) / Z
        Here we have to do the argmax first over the variable number
        of rows of each element type for each graph in the
        minibatch, then over the different types (since they are
        mutually exclusive).

        Returns:
            A list of tuples specifying the action (type, row, column).
        """

        # Uniform noise
        u = [torch.rand(i.shape, device=self.dev) for i in self.logits]
        # Gumbel noise
        gumbel = [logit - (-noise.log()).log() for logit, noise in zip(self.logits, u)]
        # scatter_max and .max create a (values, indices) pair
        # These logits are 2d (num_obj_of_type, num_actions_of_type),
        # first reduce-max over the batch, which preserves the
        # columns, so we get (minibatch_size, num_actions_of_type).
        # First we prefill `out` with very negative values in case
        # there are no corresponding logits (this can happen if e.g. a
        # graph has no edges), we don't want to accidentally take the
        # max of that type.
        mnb_max = [
            torch.zeros(self.num_graphs, i.shape[1], device=self.dev) - 1e6
            for i in self.logits
        ]
        mnb_max = [
            scatter_max(i, b, dim=0, out=out)
            for i, b, out in zip(gumbel, self.batch, mnb_max)
        ]
        # Then over cols, this gets us which col holds the max value,
        # so we get (minibatch_size,)
        col_max = [values.max(1) for values, idx in mnb_max]
        # Now we look up which row in those argmax cols was the max:
        row_pos = [
            idx_mnb[torch.arange(len(idx_col)), idx_col]
            for (_, idx_mnb), (_, idx_col) in zip(mnb_max, col_max)
        ]
        # The maxes themselves
        maxs = [values for values, idx in col_max]
        # Now we need to check which type of logit has the actual max
        type_max_val, type_max_idx = torch.stack(maxs).max(0)
        if torch.isfinite(type_max_val).logical_not_().any():
            raise ValueError(
                "Non finite max value in sample", (type_max_val, self.logits)
            )

        # Now we can return the indices of where the actions occured
        # in the form List[(type, row, column)]
        actions = []
        for i in range(type_max_idx.shape[0]):
            t = type_max_idx[i]
            # Subtract from the slice of that type and index, since the computed
            # row position is batch-wise rather graph-wise
            actions.append(
                (int(t), int(row_pos[t][i] - self.slice[t][i]), int(col_max[t][1][i]))
            )
        # It's now up to the Context class to create GraphBuildingAction instances
        # if it wants to convert these indices to env-compatible actions
        return actions

    def log_prob(self, actions: List[Tuple[int, int, int]]) -> torch.Tensor:
        """The log-probability of a list of action tuples.

        Args:
            actions: A list of action tuples (type, row, column).

        Returns:
            A tensor of shape (minibatch_size,) containing the log-probability of each action.
        """
        logprobs = self.logsoftmax()
        return torch.stack(
            [
                logprobs[t][row + self.slice[t][i], col]
                for i, (t, row, col) in enumerate(actions)
            ]
        )


class GraphBuildingEnvContext:
    """A context environment that defines what the graphs are, how they map to and from data."""

    device: str
    num_node_dim: int
    num_edge_dim: int
    num_cond_dim: int
    num_new_node_values: int
    num_node_attr_logits: int
    num_edge_attr_logits: int
    action_type_order: List[GraphActionType]

    def aidx_to_graph_action(
        self, g: gd.Data, action_idx: Tuple[int, int, int]
    ) -> GraphAction:
        """Translate an action index (e.g. from a GraphActionCategorical) to a GraphAction.

        Args:
            g: the graph to which the action is being applied.
            action_idx: the tensor indices for the corresponding action.

        Returns:
            a graph action that could be applied to the original graph coressponding to g.
        """
        raise NotImplementedError()

    def graph_action_to_aidx(
        self, g: gd.Data, action: GraphAction
    ) -> Tuple[int, int, int]:
        """Translate a GraphAction to an action index (e.g. from a GraphActionCategorical).

        Args:
            g: the graph to which the action is being applied.
            action: a graph action that could be applied to the original graph coressponding to g.

        Returns:
            the tensor indices for the corresponding action.
        """
        raise NotImplementedError()

    def graph_to_mol(self, g: Graph) -> Mol:
        """Translate a graph to a molecule.

        Args:
            g: the graph to translate.

        Returns:
            the molecule corresponding to the graph.
        """
        pass

    def sample_conditional_information(self):
        """Sample conditional information."""
        pass

    def graph_to_data(self, g: Graph) -> gd.Data:
        """Convert a networkx Graph to a torch geometric Data instance.

        Args:
            g: a graph instance.

        Returns:
            the corresponding torch_geometric graph data.
        """
        raise NotImplementedError()

    def collate(self, graphs: List[gd.Data]) -> gd.Batch:
        """Convert a list of torch geometric Data instances to a Batch
        instance.  This exists so that environment contexts can set
        custom batching attributes, e.g. by using `follow_batch`.

        Args:
            graphs: graph instances.

        Returns:
            the corresponding torch_geometric batch.
        """

        return gd.Batch.from_data_list(graphs)

    def is_sane(self, g: Graph) -> bool:
        """Verifies whether a graph is sane according to the context. This can
        catch, e.g. impossible molecules.

        Args:
            g: a graph.

        Returns:
            true if the environment considers g to be sane.
        """

        raise NotImplementedError()

    def mol_to_graph(self, mol: Mol) -> Graph:
        """Verifies whether a graph is sane according to the context. This can
        catch, e.g. impossible molecules.

        Args:
            mol: an RDKit molecule.

        Returns:
            the corresponding Graph representation of that molecule.
        """

        raise NotImplementedError()



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Implementation of PaccMann^GP conditional generator."""

import json
import logging
import os
from copy import deepcopy
from typing import Any, Dict, List, Optional, Union

import torch
from paccmann_chemistry.models.vae import StackGRUDecoder, StackGRUEncoder, TeacherVAE
from paccmann_chemistry.utils.search import SamplingSearch
from paccmann_gp.affinity_minimization import AffinityMinimization
from paccmann_gp.callable_minimization import CallableMinimization
from paccmann_gp.combined_minimization import CombinedMinimization
from paccmann_gp.gp_optimizer import GPOptimizer
from paccmann_gp.mw_minimization import MWMinimization
from paccmann_gp.qed_minimization import QEDMinimization
from paccmann_gp.sa_minimization import SAMinimization
from paccmann_gp.smiles_generator import SmilesGenerator
from paccmann_predictor.models import MODEL_FACTORY
from pytoda.proteins.protein_language import ProteinLanguage
from pytoda.smiles.smiles_language import SMILESLanguage

from ....frameworks.torch import device_claim

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

MINIMIZATION_FUNCTIONS = {
    "qed": QEDMinimization,
    "sa": SAMinimization,
    "molwt": MWMinimization,
    "affinity": AffinityMinimization,
    "callable": CallableMinimization,
}


class GPConditionalGenerator:
    """Conditional generator as implemented in https://doi.org/10.1021/acs.jcim.1c00889."""

    def __init__(
        self,
        resources_path: str,
        temperature: float = 1.4,
        generated_length: int = 100,
        batch_size: int = 32,
        limit: float = 5.0,
        acquisition_function: str = "EI",
        number_of_steps: int = 32,
        number_of_initial_points: int = 16,
        initial_point_generator: str = "random",
        seed: int = 42,
        number_of_optimization_rounds: int = 1,
        sampling_variance: float = 0.1,
        samples_for_evaluation: int = 4,
        maximum_number_of_sampling_steps: int = 32,
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """Initialize the conditional generator.

        Args:
            resources_path: directory where to find models and parameters.
            temperature: temperature parameter for the softmax sampling in decoding. Defaults to 1.4.
            generated_length: maximum length in tokens of the generated molcules (relates to the SMILES length). Defaults to 100.
            batch_size: batch size used for the generative model sampling. Defaults to 16.
            limit: hypercube limits in the latent space. Defaults to 5.0.
            acquisition_function: acquisition function used in the Gaussian process. Defaults to "EI". More details in https://scikit-optimize.github.io/stable/modules/generated/skopt.gp_minimize.html.
            number_of_steps: number of steps for an optmization round. Defaults to 32.
            number_of_initial_points: number of initial points evaluated. Defaults to 16.
            initial_point_generator: scheme to generate initial points. Defaults to "random". More details in https://scikit-optimize.github.io/stable/modules/generated/skopt.gp_minimize.html.
            seed: seed used for random number generation in the optimizer. Defaults to 42.
            number_of_optimization_rounds: maximum number of optimization rounds. Defaults to 1.
            sampling_variance: variance of the Gaussian noise applied during sampling from the optimal point. Defaults to 0.1.
            samples_for_evaluation: number of samples averaged for each minimization function evaluation. Defaults to 4.
            maximum_number_of_sampling_steps: maximum number of sampling steps in an optmization round. Defaults to 32.
            device: . Defaults to None, a.k.a, picking a default one ("gpu" if present, "cpu" otherwise).
        """
        # device
        self.device = device_claim(device)
        # setting sampling parameters
        self.temperature = temperature
        self.generated_length = generated_length
        self.batch_size = batch_size
        # setting VAE parameters
        self.svae_params = dict()
        with open(os.path.join(resources_path, "vae_model_params.json"), "r") as f:
            self.svae_params.update(json.load(f))
        smiles_language = SMILESLanguage.load(
            os.path.join(resources_path, "selfies_language.pkl")
        )
        # initialize encoder, decoder, testVAE, and GP_generator_MW
        self.gru_encoder = StackGRUEncoder(self.svae_params).to(self.device)
        self.gru_decoder = StackGRUDecoder(self.svae_params).to(self.device)
        self.gru_vae = TeacherVAE(self.gru_encoder, self.gru_decoder)
        self.gru_vae.load_state_dict(
            torch.load(
                os.path.join(resources_path, "vae_weights.pt"),
                map_location=self.device,
            )
        )
        self.gru_vae._associate_language(smiles_language)
        self.gru_vae.eval()
        self.smiles_generator = SmilesGenerator(
            self.gru_vae,
            search=SamplingSearch(temperature=self.temperature),
            generated_length=self.generated_length,
        )
        self.latent_dim = self.gru_decoder.latent_dim
        # setting affinity predictor parameters
        with open(os.path.join(resources_path, "mca_model_params.json")) as f:
            self.predictor_params = json.load(f)
        self.affinity_predictor = MODEL_FACTORY["bimodal_mca"](
            self.predictor_params
        ).to(self.device)
        self.affinity_predictor.load(
            os.path.join(resources_path, "mca_weights.pt"),
            map_location=self.device,
        )
        affinity_protein_language = ProteinLanguage.load(
            os.path.join(resources_path, "protein_language.pkl")
        )
        affinity_smiles_language = SMILESLanguage.load(
            os.path.join(resources_path, "smiles_language.pkl")
        )
        self.affinity_predictor._associate_language(affinity_smiles_language)
        self.affinity_predictor._associate_language(affinity_protein_language)
        self.affinity_predictor.eval()
        # setting optimizer parameters
        self.limit = limit
        self.acquisition_function = acquisition_function
        self.number_of_initial_points = number_of_initial_points
        if number_of_steps < self.number_of_initial_points:
            logger.warning(
                "number of initial points is larger than number of steps "
                f"({self.number_of_initial_points}/{number_of_steps}). "
                f"Resetting number of steps to {self.number_of_initial_points}."
            )
            self.number_of_steps = self.number_of_initial_points
        else:
            self.number_of_steps = number_of_steps
        self.initial_point_generator = initial_point_generator
        self.seed = seed
        self.number_of_optimization_rounds = number_of_optimization_rounds
        self.sampling_variance = sampling_variance
        self.samples_for_evaluation = samples_for_evaluation
        self.maximum_number_of_sampling_steps = maximum_number_of_sampling_steps

    def target_to_minimization_function(
        self, target: Union[Dict[str, Dict[str, Any]], str]
    ) -> CombinedMinimization:
        """Use the target to configure a minimization function.

        Args:
            target: dictionary or JSON string describing the optimization target.

        Returns:
            a minimization function.
        """
        if isinstance(target, str):
            target_dictionary = json.loads(target)
        elif isinstance(target, dict):
            target_dictionary = deepcopy(target)
        else:
            raise ValueError(
                f"{target} of type {type(target)} is not supported: provide 'str' or 'Dict[str, Dict[str, Any]]'"
            )
        minimization_functions = []
        weights = []
        for minimization_function_name, parameters in target_dictionary.items():
            weight = 1.0
            if "weight" in parameters:
                weight = parameters.pop("weight")
            function_parameters = {
                **parameters,
                **{
                    "batch_size": self.samples_for_evaluation,
                    "smiles_decoder": self.smiles_generator,
                },
            }
            minimization_function = MINIMIZATION_FUNCTIONS[minimization_function_name]
            if minimization_function_name == "affinity":
                function_parameters["affinity_predictor"] = self.affinity_predictor
            minimization_functions.append(minimization_function(**function_parameters))
            weights.append(weight)
        return CombinedMinimization(
            minimization_functions=minimization_functions,
            batch_size=1,
            function_weights=weights,
        )

    def generate_batch(self, target: Any) -> List[str]:
        """Generate molecules given a target.

        Args:
            target: dictionary or JSON string describing the optimization target.

        Returns:
            a list of molecules as SMILES string.
        """
        # make sure the seed is transformed to avoid redundancy over multiple calls (using Knuth multiplicative hashing)
        self.seed = self.seed * 2654435761 % 2**32
        logger.info(f"configuring optimization for target: {target}")
        # target configuration
        self.target = target
        self.minimization_function = self.target_to_minimization_function(self.target)
        # optimizer configuration
        self.target_optimizer = GPOptimizer(self.minimization_function.evaluate)
        optimization_parameters = dict(
            dimensions=[(-self.limit, self.limit)] * self.latent_dim,
            acq_func=self.acquisition_function,
            n_calls=self.number_of_steps,
            n_initial_points=self.number_of_initial_points,
            initial_point_generator=self.initial_point_generator,
            random_state=self.seed,
        )
        logger.info(
            f"running optimization with the following parameters: {optimization_parameters}"
        )
        smiles_set = set()
        logger.info(
            f"running at most {self.number_of_optimization_rounds} optmization rounds"
        )
        for optimization_round in range(self.number_of_optimization_rounds):
            logger.info(f"starting round {optimization_round + 1}")
            optimization_parameters["random_state"] += optimization_round  # type:ignore
            res = self.target_optimizer.optimize(optimization_parameters)
            latent_point = torch.tensor([[res.x]])
            smiles_set_per_round = set()

            logger.info(f"starting sampling for {optimization_round + 1}")
            for _ in range(self.maximum_number_of_sampling_steps):
                generated_smiles = self.smiles_generator.generate_smiles(
                    latent_point.repeat(1, self.batch_size, 1)
                    + torch.cat(
                        (
                            torch.zeros(1, 1, self.latent_dim),
                            (self.sampling_variance**0.5)
                            * torch.randn(1, self.batch_size - 1, self.latent_dim),
                        ),
                        dim=1,
                    )
                )
                smiles_set_per_round.update(set(generated_smiles))
            smiles_set.update(smiles_set_per_round)
            logger.info(f"completing round {optimization_round + 1}")
        logger.info(f"generated {len(smiles_set)} molecules in the current run")
        return list(
            [molecule_smiles for molecule_smiles in smiles_set if molecule_smiles]
        )



#
# MIT License
#
# Copyright (c) 2023 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Implementation of Regression Transformer conditional generators."""
import json
import logging
import os
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import numpy as np
import torch
from rdkit import Chem
from terminator.collators import MaskedTextCollator, PropertyCollator
from terminator.inference import InferenceRT
from terminator.search import SEARCH_FACTORY, Search
from terminator.selfies import decoder, encoder
from terminator.tokenization import InferenceBertTokenizer, PolymerGraphTokenizer
from transformers import AutoConfig, AutoModelWithLMHead, XLNetLMHeadModel

from ....domains.materials import MoleculeFormat, Sequence, validate_molecules
from ....frameworks.torch import device_claim, map_tensor_dict
from .utils import filter_stubbed, get_substructure_indices

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ConditionalGenerator:
    """Main interface for a regression transformer."""

    # device where the inference is running.
    device: torch.device

    # The task the RT is currently performing. Either 'regression' or 'generation'.
    task: str

    # method to convert logits to predictions. Either GreedySearch or SamplingSearch.
    search: Search

    # ***** Additional attributes for text generation *****
    # data collator for property prediction of self-generated items.
    property_collator: PropertyCollator

    # The input format for the substructures
    subs_format: str

    # number of samples obtained per call.
    batch_size: int = 8

    # a dictionary to specify sampling from a specific seed molecule.
    sampling_wrapper: Dict[str, Any] = {}

    def __init__(
        self,
        resources_path: str,
        device: Optional[Union[torch.device, str]] = None,
        tolerance: Union[float, Dict[str, float]] = 20.0,
    ) -> None:
        """
        Initialize the generator.

        Args:
            resources_path: directory where to find models and parameters.
            device: device where the inference is running either as a dedicated class
                or a string. If not provided is inferred.
            tolerance:  percentage of tolerated deviation between desired and obtained property.
                Either a single float or a dictionary of floats, where the keys are the properties.
                Note that the tolerance is *only* used for post-hoc filtering of the generated molecules.
        """
        # device
        self.device = device_claim(device)
        # Set up the data preparation pipeline
        if not os.path.exists(os.path.join(resources_path, "inference.json")):
            raise OSError(
                f"algorithm_version {resources_path.split('/')[-1]} does not exist."
            )
        self.tokenizer = InferenceBertTokenizer.from_pretrained(
            resources_path, pad_even=False
        )
        self.collator = MaskedTextCollator(self.tokenizer)

        # Set up model: First load the pretrained XLNet model
        xlnet_model, config = self.load_model(resources_path)
        # Initialize the custom RT model
        self.model = InferenceRT(xlnet_model, self.tokenizer, config)

        # Set up inference parameters
        self.tolerance = tolerance
        self.load_inference(resources_path)

    def load_model(self, resources_path: str) -> Tuple[XLNetLMHeadModel, Any]:
        """
        Loading a XLNetLMHeadModel which constitutes the base of a RT model.

        Args:
            resources_path: path to the model.

        Returns:
            XLNetLMHeadModel: base of a Regression Transformer model.
            XLNetConfig: configuration of the model.
        """
        config_name = os.path.join(resources_path, "config.json")
        config = AutoConfig.from_pretrained(config_name, mem_len=1024)
        xlnet_model = AutoModelWithLMHead.from_pretrained(
            resources_path, from_tf=False, config=config
        )
        xlnet_model.resize_token_embeddings(len(self.tokenizer))
        xlnet_model.to(self.device)
        xlnet_model.eval()
        logger.info(f"Model restored from {resources_path}")
        return xlnet_model, config

    def load_inference(self, resources_path: str) -> None:
        """
        Load and set up all parameters necessary for inference.

        Args:
            resources_path: path to the model folder.
        """
        try:
            with open(os.path.join(resources_path, "inference.json"), "r") as f:
                data = json.load(f)
            self.properties = data["property_token"]
            if isinstance(self.properties, str):
                self.properties = [self.properties]

            # Optional normalize parameter (for property) defaults to True
            self.do_normalize = data.get("normalize", [True] * len(self.properties))

            self.property_mask_lengths = [
                data["property_mask_length"][p] for p in self.properties
            ]
            self._mins = [
                data.get("property_ranges", {}).get(p, [0, 1])[0]
                for p in self.properties
            ]
            self._maxs = [
                data.get("property_ranges", {}).get(p, [0, 1])[1]
                for p in self.properties
            ]
            self.metadata = data

            # If tolerance dict is given, ensure it is well-formed
            self.tolerance_dict = {}
            if isinstance(self.tolerance, Dict):
                # Check that no extra properties are given
                for key in self.tolerance.keys():
                    if key not in self.properties:
                        logger.error(
                            f"tolerance key {key} is not a valid property name, will be ignored."
                        )
                # Check that all properties are given
                for key in self.properties:
                    if key not in self.tolerance.keys():
                        logger.error(
                            f"tolerance key {key} is missing, will be set to 20%."
                        )
                    self.tolerance_dict[key] = self.tolerance.get(key, 20.0)

            elif isinstance(self.tolerance, float):
                for key in self.properties:
                    self.tolerance_dict[key] = self.tolerance
            else:
                raise TypeError(
                    f"tolerance must be either a float or a dictionary, not {type(self.tolerance)}."
                )

            # Tolerance defined on the original scale
            self.tolerances = [
                (self._maxs[i] - self._mins[i]) * self.tolerance_dict[k] / 100.0
                for i, k in enumerate(self.properties)
            ]
        except Exception:
            raise ValueError(
                f"Could not restore inference parameters from {resources_path}"
            )

    def denormalize(self, x: float, idx: int, precision: int = 4) -> float:
        """
        Denormalize from the model scale to the original scale.

        Args:
            x: normalized value (often in [0,1]).
            idx: index of the property.
            precision: optional rounding precision. Defaults to 4.

        Returns:
            float: Value in regular scale.
        """

        # If the property was not normalized, return the value
        if not self.do_normalize[idx]:
            return x

        return round(
            x * (self._maxs[idx] - self._mins[idx]) + self._mins[idx], precision
        )

    def normalize(self, x: str, idx: int, precision: int = 3) -> float:
        """
        Normalize from original scale to desired scale.

        Args:
            x: unnormalized input.
            idx: index of the property.
            precision: optional rounding precision.

        Returns:
            float: Normalized value.
        """
        # Error handling
        if not isinstance(x, float) and not self.isfloat(x):
            raise TypeError(f"{x} is not a float and cant safely be casted.")

        x_float = float(x)
        # If this property does not require normalization, return it
        if not self.do_normalize[idx]:
            return x_float
        normed = round(
            (x_float - self._mins[idx]) / (self._maxs[idx] - self._mins[idx]), precision
        )
        return normed

    def validate_input(self, x: str) -> None:
        """
        Sanity checking for formatting of the input string.

        Args:
            x: The string to be validated.

        Raises:
            ValueError: If string was formatted incorrectly.
        """

        if self.tokenizer.expression_separator not in x:
            raise ValueError(
                f"Expression separator {self.tokenizer.expression_separator} not "
                f"found in input {x}."
            )
        if self.tokenizer.mask_token not in x:
            raise ValueError(
                f"Nothing to do, no mask to fill ({self.tokenizer.mask_token}) found "
                f"in input {x}."
            )
        if not all(p in x for p in self.properties):
            raise ValueError(
                f"Not all property tokens ({self.properties}) found in input"
            )

        text_sequence = x.split(self.tokenizer.expression_separator)[-1]
        number_sequence = x[: -len(text_sequence) - 1]

        if (
            self.tokenizer.mask_token in text_sequence
            and self.tokenizer.mask_token in number_sequence
        ):
            raise ValueError(
                f"Do not mask number and text sequence at the same time like in {x}."
            )
        if isinstance(self.tokenizer.text_tokenizer, PolymerGraphTokenizer):
            self.validate_input_molecule(text_sequence, MoleculeFormat.copolymer)
        else:
            # We can assume this to be a SELFIES
            self.validate_input_molecule(text_sequence)

        self.validate_input_numerical(number_sequence)

    def validate_input_molecule(
        self, sequence: str, input_type: str = "SELFIES"
    ) -> None:
        """
        Verifies that the non-numerical part of the input is a proper sequence.

        Args:
            sequence: input sequence to be validated.
        """
        raise NotImplementedError

    def validate_input_numerical(self, sequence: str) -> None:
        """
        Verifies that the numeric part of the input sequence is valid.

        Args:
            sequence: input sequence to be validated.
        """
        if self.tokenizer.mask_token in sequence:
            # Task is most likely regression
            if sequence.count(self.tokenizer.mask_token) != sum(
                self.property_mask_lengths
            ):
                raise ValueError(
                    f"To predict {self.properties} you have to mask exactly "
                    f"{self.property_mask_lengths} times respectively."
                )
        else:
            # Task is most likely generation
            tokens = []
            for toks in sequence.split(self.tokenizer.expression_separator):
                tokens.extend(self.tokenizer.property_tokenizer.tokenize(toks))
                tokens.append(self.tokenizer.expression_separator)
            idx = list(
                np.where(np.array(tokens) == self.tokenizer.expression_separator)[0]
            )
            if len(idx) < len(self.properties):
                raise ValueError(
                    "Please append all properties with separator token:"
                    f"{self.tokenizer.expression_separator}."
                )
            if idx[0] != self.property_mask_lengths[0] + 1:
                raise ValueError(
                    f"To generate samples, please describe {self.properties[0]} with "
                    f"{self.property_mask_lengths[0]} tokens."
                )
            if len(self.properties) > 1:
                for i in range(len(self.properties) - 1):
                    if idx[i + 1] - idx[i] - 2 != self.property_mask_lengths[i + 1]:
                        raise ValueError(
                            f"To generate samples, describe {self.properties[i + 1]}"
                            f" with {self.property_mask_lengths[i+1]} samples."
                        )

    def safely_determine_task(self, x: str) -> str:
        """
        Determines whether the passed sequence adheres to regression or generation task.

        Args:
            x: the user-provided input sequence for the model, inluding mask tokens.

        Raises:
            ValueError: if the sequence does not adhere to the formatting rules.

        Returns:
            str: the task, either 'regression' or 'generation'.
        """

        self.validate_input(x)
        if (
            self.tokenizer.mask_token
            in x.split(self.tokenizer.expression_separator)[-1]
        ):

            return "generation"

        return "regression"

    def generate_batch_regression(self, context: Sequence) -> List[Sequence]:
        """
        Predict the property of a sample.

        Args:
            context: a string with a masked property, a separator and an
                entity. E.g. <stab>[MASK][MASK][MASK][MASK]|GSQEVNSGTQTYKNASPEEAERIARKAGATTWTEKGNKWEIRI.

        Returns:
            List[Sequence]: a list of (denormalized) predicted properties for the entity.
                Stored as a Sequence (str), e.g., '<qed>0.727'.
        """
        logger.info(f"Starting prediction for sequence {context}")

        # Prepare the batch
        inputs = self.collator([self.tokenizer(context)])
        input_ids = inputs["input_ids"].cpu()

        # Forward pass
        outputs = self.model(map_tensor_dict(inputs, self.device))

        # Obtain the singular predictions
        prediction = self.search(outputs["logits"].detach())

        return self.compile_regression_result(input_ids, prediction)

    def compile_regression_result(
        self, input_ids: torch.Tensor, prediction: torch.Tensor
    ) -> List[Sequence]:
        """
        Postprocesses the prediction from the property task to obtain a float.

        Args:
            input_ids: 2D Tensor of shape (batch_size, sequence_length).
            prediction: 2D Tensor of shape (batch_size, sequence_length).

        Returns:
            List[Sequence]: list of property sequences (one per sample). Can contain
                multiple properties but have to be hashable, therefore we use Sequences,
                e.g., '<qed>0.727' or '<logp>6.65<scscore>3.82'.
        """
        properties = []
        for inp, pred in zip(input_ids, prediction):
            in_tokens = self.tokenizer.decode(
                inp, clean_up_tokenization_spaces=False
            ).split(" ")
            out_tokens = self.tokenizer.decode(
                pred, clean_up_tokenization_spaces=False
            ).split(" ")
            joined = self.tokenizer.get_sample_prediction(out_tokens, in_tokens)
            _, gen_prop = self.tokenizer.aggregate_tokens(joined, label_mode=False)
            properties.append(
                "".join(
                    [
                        f"<{k}>{self.denormalize(v, pidx)}"
                        for pidx, (k, v) in enumerate(gen_prop.items())
                    ]
                )
            )
        return properties

    def generate_batch_generation(self, sequence: Sequence) -> Tuple:
        """
        Conditionally generate sequences given a continuous property value and a fixed
        sequence. This function first conditionally generates the novel sequences and
        then predicts their properties using the RT again. Only if the predicted
        property is within the tolerance range, the novel sequence is returned.

        Args:
            sequence: the input sequence with masked tokens on the text.

        Returns:
            Tuple[Tuple[str, float]]: a tuple of tuples, each containing the generated
                sequence alongside its predicted property value.
        """
        self.target = sequence

        # If we are in sampling mode, the sampled sequence changes per iteration.
        if self.sampling_wrapper != {}:
            sequence = self.sample_sequence(sequence)

        # The property score has to be in the range understood by the model
        sequence = self.normalize_sequence(sequence)

        logger.info(f"Starting prediction for sequence {sequence}")

        # Prepare the batch
        tokens = self.tokenizer(sequence)
        inputs = self.collator([tokens] * self.batch_size)
        input_ids = inputs["input_ids"].clone()

        # Forward pass
        outputs = self.model(map_tensor_dict(inputs, self.device))
        # Obtain model predictions via the search method
        predictions = self.search(outputs["logits"].detach()).squeeze().cpu()
        # Combine predictions with the static part to obtain the full sequences
        generations = input_ids.cpu()
        generations[generations == self.tokenizer.mask_token_id] = predictions[
            generations == self.tokenizer.mask_token_id
        ]

        # Second part: Predict the properties of the just generated sequence
        _input = self.property_collator.mask_tokens(generations)
        prediction_input = {
            "input_ids": _input[0],
            "perm_mask": _input[1],
            "target_mapping": _input[2],
            "attention_mask": self.property_collator.attention_mask(generations),
        }

        # Pass through model
        property_outputs = self.model(map_tensor_dict(prediction_input, self.device))
        # It's a design choice to go with greedy predictions here
        predictions = torch.argmax(property_outputs["logits"].detach(), dim=-1)
        # Obtain floating predictions
        properties = self.compile_regression_result(generations, predictions)
        # Obtain the sequences (AAS or SELFIES)
        sequences = [
            self.tokenizer.to_readable(
                "".join(
                    self.tokenizer.decode(seq, skip_special_tokens=True).split(" ")
                ).split(self.tokenizer.expression_separator)[-1]
            )
            for seq in generations
        ]

        # Filter out all sequences that do not satisfy property constraints within
        # tolerance range.
        logger.debug(f"Sequences {sequences}, properties: {properties}")
        property_successes: Tuple = tuple(
            filter(
                lambda x: all(
                    [
                        abs(
                            float(x[1].split(p)[-1].split("<")[0])
                            - self.target_values[i]
                        )
                        < self.tolerances[i]
                        for i, p in enumerate(self.properties)
                    ]
                ),
                zip(sequences, properties),
            )
        )  # type: ignore

        # filter out sequences that do not include desired substructures
        successes = self.filter_substructures(property_successes)  # type: ignore
        logger.info(f"Successes: {successes}")
        return successes

    def normalize_sequence(self, context: Sequence) -> Sequence:
        """
        Take a sequence with unnormalized property score(s) and convert it to a
        sequence with a normalized score.

        Args:
            context: sequence with unnormalized property.

        Returns:
            sequence with normalized property.
        """

        # Tokenize sequence and extract positions of separator tokens
        tokens = self.tokenizer.tokenize(context)
        final_tokens = ""
        sep_idxs = [
            i for i, t in enumerate(tokens) if t == self.tokenizer.expression_separator
        ]

        # Declard as class variable since used by other methods
        self.target_values = []

        # Loop over properties and normalize them
        for idx, prop in enumerate(self.properties):
            numerical_tokens = tokens[tokens.index(prop) + 1 : sep_idxs[idx]]
            final_tokens += prop

            unnorm = self.tokenizer.floating_tokens_to_float(numerical_tokens)
            self.target_values.append(unnorm)
            target = self.normalize(unnorm, idx)
            norm = str(target + 1e-10)[: self.property_mask_lengths[idx]]
            final_tokens += norm + self.tokenizer.expression_separator

        # Append rest of sequence
        final_tokens += "".join(tokens[sep_idxs[-1] + 1 :])
        return final_tokens

    @staticmethod
    def isfloat(sequence: str) -> bool:
        """Safely determine whether a string can be converted to a float

        Args:
            sequence: A string

        Returns:
            Whether it can be converted to a float
        """
        try:
            float(sequence)
            return True
        except ValueError:
            return False

    def validate_numerical(
        self, sequences: List[Any]
    ) -> Tuple[List[Sequence], List[int]]:
        """
        Validate whether a list of sequences contains only numerical values.

        Args:
            sequences: a list of hopefully only numerical values.

        Returns:
            A tuple of two lists for the validated Sequences and their respective
            indices.
        """
        items = []
        idxs = []
        for idx, item in enumerate(sequences):
            if (
                isinstance(item, str)
                and item.startswith("<")
                and self.isfloat(item.split(">")[-1])
            ):
                items.append(item)
                idxs.append(idx)
        return items, idxs

    def validate_sampling_wrapper(
        self,
        context: str,
        property_goal: Dict[str, Any] = {},
        fraction_to_mask: float = 0.2,
        tokens_to_mask: List = [],
        substructures_to_mask: List[str] = [],
        substructures_to_keep: List[str] = [],
        text_filtering: bool = False,
    ) -> None:
        """
        Validating whether the wrapper can be used for conditional generation of samples.

        Args:
            context: A string that is used as a seed. Has to be a SMILES or a block-copoylmer
                (RegressionTransformerMolecules) or AAS (RegressionTransformerProteins).
            property_goal: Specifies the property conditions for the targeted generation.
               The keys are the properties and have to be aligned with the
               algorithm_version. For example, for the solubility model use:
               {'<esol>': 1.23}
               or for the logp_and_synthesizability model use:
               {'<logp>': 1.23, '<synthesizability>': 2.34}
                Defaults to {}, but it has to be specified.
            fraction_to_mask: The fraction of tokens that can be changed. Defaults to 0.2.
            tokens_to_mask: A list of atoms (or amino acids) that can be considered for masking.
                Defaults to [] meaning that all tokens can be masked. E.g., use ['F'] to
                only mask fluorine atoms.
            substructures_to_mask: Specifies a list of substructures that should be masked.
                Given in SMILES format. This is excluded from the stochastic masking.
                NOTE: The model operates on SELFIES and the matching of the substructures occurs
                in SELFIES simply on a string level.
            substructures_to_keep: Specifies a list of substructures that should definitely be kept.
                Given in SMILES format. This is excluded from the stochastic masking.
                NOTE: This keeps tokens even if they are included in `tokens_to_mask`.
                NOTE: The model operates on SELFIES and the matching of the substructures occurs
                in SELFIES simply on a string level.
            text_filtering: Generated sequences are post-hoc filtered for the presence of
                `substructures_to_keep`. This is done with RDKit substructure matches. If the sub-
                structure cant be converted to a mol object, this argument toggles whether a substructure
                should be ignored from post-hoc filtering (this happens per default) or whether
                filtering should occur on a pure string level.
                NOTE: This does not affect the actual generation process.
                Defaults to False.
        """
        if isinstance(self.tokenizer.text_tokenizer, PolymerGraphTokenizer):
            self.validate_input_molecule(context, MoleculeFormat.copolymer)
        else:
            self.validate_input_molecule(context, MoleculeFormat.smiles)

        self.seed_molecule = context

        if property_goal == {}:
            raise ValueError("Please specify the target properties with a dictionary.")
        if not all([k in property_goal.keys() for k in self.properties]):
            raise ValueError(
                f"Please provide property goals for all properties: {self.properties}."
            )
        self.property_goal = property_goal

        if not isinstance(fraction_to_mask, float):
            raise TypeError(
                f"The fraction_to_mask {fraction_to_mask} has to be a float."
            )
        if fraction_to_mask < 0 or fraction_to_mask > 1:
            raise ValueError(
                f"The fraction_to_mask {fraction_to_mask} has to be between 0 and 1."
            )
        self.fraction_to_mask = fraction_to_mask

        if not isinstance(tokens_to_mask, list):
            raise TypeError(f"The tokens_to_mask {tokens_to_mask} has to be a list.")
        self.maskable_tokens = self.get_maskable_tokens(tokens_to_mask.copy())

        found_maskable = list(filter(lambda x: x in context, tokens_to_mask))
        if len(found_maskable) == 0 and tokens_to_mask != []:
            raise ValueError(
                "Malformed search query: None of the `tokens_to_mask` "
                f"{tokens_to_mask} are in the `context`: {context}."
            )
        elif len(found_maskable) < len(tokens_to_mask):
            logger.warning(
                "Malformed search query: Some of the `tokens_to_mask` "
                f"{tokens_to_mask} are not in the `context`: {context}."
            )

        if not isinstance(text_filtering, bool):
            raise TypeError(f"The text_filtering {text_filtering} has to be a bool.")
        self.text_filtering = text_filtering

        self.validate_substructures(substructures_to_mask, substructures_to_keep)

        logger.info(
            f"Will start sampling molecules similar to {context} with goal: "
            f"{property_goal} and masking {fraction_to_mask} of the tokens."
        )

    def sample_sequence(self, seq: str) -> str:
        """
        Assembling a RT-sequence from a seed SMILES/AAS sequence.

        Args:
            seq: A SMILES/AAS string used as seed.

        Returns:
            A RT-sequence that uses SELFIES/AAS and incorporates the properties, e.g.:
            `<logp>1.234|<synthesizability>1.234|[C][C][O]`
        """

        if self.sampling_wrapper == {}:
            return seq

        # Convert SMILES/AAS to list of SELFIES/AAS tokens
        language_seq = self.language_encoding(seq)
        tokens = self.tokenizer.text_tokenizer.tokenize(language_seq)

        # Determine which tokens can be masked (from `tokens_to_mask`)
        maskable_tokens = (
            set(tokens) if self.maskable_tokens == [] else set(self.maskable_tokens)
        )
        maskable_idxs = [i for i, t in enumerate(tokens) if t in maskable_tokens]

        # Make sure that `substructures_to_keep` are not masked
        for keep in self.substructures_to_keep:
            keep_toks = self.tokenizer.text_tokenizer.tokenize(
                self.language_encoding(keep)
            )
            to_be_kept = get_substructure_indices(tokens, keep_toks)
            maskable_idxs = list(filter(lambda x: x not in to_be_kept, maskable_idxs))

        num_to_mask = round(len(maskable_idxs) * self.fraction_to_mask)
        # Mask the tokens
        mask_idxs = np.random.choice(maskable_idxs, num_to_mask, replace=False)
        sequence_tokens = [
            t if i not in mask_idxs else self.tokenizer.mask_token
            for i, t in enumerate(tokens)
        ]

        # Make sure that `substructures_to_mask` are actually masked
        for mask in self.substructures_to_mask:
            mask_toks = self.tokenizer.text_tokenizer.tokenize(
                self.language_encoding(mask)
            )
            to_be_masked = get_substructure_indices(tokens, mask_toks)
            for idx in to_be_masked:
                sequence_tokens[idx] = self.tokenizer.mask_token

        # Extract the property tokens
        if not set(self.properties) == set(self.property_goal.keys()):
            raise ValueError(
                f"Please provide property goals exactly for: {self.properties}."
            )

        # Define property tokens (this has unnormalized property values)
        prop_tokens = ""
        for i, p in enumerate(self.properties):
            numerical = str(self.property_goal[p] + 1e-10)[
                : self.property_mask_lengths[i]
            ]
            prop_tokens += p + numerical + self.tokenizer.expression_separator

        # Return new sequence
        sequence = prop_tokens + "".join(sequence_tokens)
        return sequence

    def get_maskable_tokens(self, tokens_to_mask: List[str]):
        raise NotImplementedError

    def language_encoding(self, seq: str):
        raise NotImplementedError

    def filter_substructures(self, property_successes: Tuple[Tuple[str, str]]):
        raise NotImplementedError

    def validate_substructures(
        self, substructures_to_mask: List[str], substructures_to_keep: List[str]
    ):
        """
        Validates the substructures that are ignored/kept for the masking when the
        `sampling_wrapper` is used.


        Args:
            substructures_to_mask: List of substructures that should be masked.
            substructures_to_keep: List of substructures that should be kept.

        Raises:
            NotImplementedError: Implemented by the child classes.
        """
        seed_encoded = self.language_encoding(self.seed_molecule)
        if not isinstance(substructures_to_mask, list):
            raise TypeError(
                f"The substructures_to_mask {substructures_to_mask} has to be a list."
            )
        to_mask = []
        for mask in substructures_to_mask:
            if not isinstance(mask, str):
                raise TypeError(
                    f"The substructure_to_mask {mask} has to be a string in {self.subs_format} format."
                )
            if self.language_encoding(mask) not in seed_encoded:
                logger.error(
                    f"\nSubstructure to keep in {mask} not found in seed sequence, ignoring it."
                )
            else:
                to_mask.append(mask)
        self.substructures_to_mask = to_mask

        if not isinstance(substructures_to_keep, list):
            raise TypeError(
                f"The substructures_to_keep {substructures_to_keep} has to be a list."
            )
        to_keep = []
        for keep in substructures_to_keep:
            if not isinstance(keep, str):
                raise TypeError(
                    f"The substructure_to_keep {keep} has to be a string in AAS format."
                )
            if self.language_encoding(keep) not in seed_encoded:
                logger.error(
                    f"\nSubstructure to keep in {keep} not found in seed sequence, ignoring it."
                )
            else:
                to_keep.append(keep)

        self.substructures_to_keep = to_keep
        if len(set(self.substructures_to_keep + self.substructures_to_mask)) < len(
            self.substructures_to_keep
        ) + len(self.substructures_to_mask):
            raise ValueError(
                "Substructures to mask and keep contain duplicates or overlap."
            )


class ChemicalLanguageRT(ConditionalGenerator):
    """
    Hybrid regression and conditional molecular generation model as implemented in
    https://arxiv.org/abs/2202.01338.

    Attributes:
        resources_path: path to the model.
        context: user-specified input text for the model.
        search: search key to instantiate a search via terminator.search.SEARCH_FACTORY.
        temperature: the temperature parameter in case of a `sample` search.
        batch_size: the batch size for the model, applicable only to generative task.
        tolerance: the tolerance for the property of the generated molecules.
    """

    subs_format: str = "SMILES"

    def __init__(
        self,
        resources_path: str,
        context: str,
        search: str = "sample",
        temperature: float = 1.4,
        batch_size: int = 8,
        tolerance: Union[float, Dict[str, float]] = 20.0,
        sampling_wrapper: Dict[str, Any] = {},
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """
        Initialize the molecule generator.

        Args:
            resources_path: directory where to find models and parameters.
            context: user-specified input text for the model.
            search: search key to instantiate a search, defaults to `sample`.
            temperature: temperature for the sampling. Defaults to 1.4.
            batch_size: number of points sampled per call. Defaults to 8.
            tolerance: the tolerance for the property of the generated molecules.
                Given in percent. Defaults to 20.
            sampling_wrapper: A high-level entry point that allows specifying a seed
                SMILES alongside some target conditions.
                NOTE: If this is used, the `target` needs to be a single SMILES string.
                Example: {
                    'fraction_to_mask': 0.5,
                    'tokens_to_mask': [],
                    'property_goal': {'<qed>': 0.85}
                }
            device: device where the inference s running either as a dedicated class
                or a string. If not provided is inferred.
        """
        super().__init__(
            device=device, resources_path=resources_path, tolerance=tolerance
        )

        if sampling_wrapper == {}:
            # Validate input and determine task
            self.task = self.safely_determine_task(context)
        else:
            self.validate_sampling_wrapper(context=context, **sampling_wrapper)
            self.task = "generation"
        self.sampling_wrapper = sampling_wrapper

        # Console outputs for usage of search methods
        if search == "sample" and self.task == "regression":
            logger.warning("For regression task, greedy search is recommended")
        elif search == "greedy" and self.task == "generation":
            logger.warning("For generation task, sample search is recommended")

        if search not in SEARCH_FACTORY.keys():
            raise KeyError(f"Pick a search of {SEARCH_FACTORY.keys()} not: {search}.")
        self.search = SEARCH_FACTORY[search](temperature=temperature)

        # Based on task, set the correct generate_batch method
        if self.task == "regression":
            self.generate_batch = self.generate_batch_regression
        elif self.task == "generation":
            self.generate_batch = self.generate_batch_generation  # type: ignore
            self.batch_size = batch_size
            self.property_collator = PropertyCollator(
                tokenizer=self.tokenizer,
                property_tokens=self.properties,
                num_tokens_to_mask=[-1] * len(self.properties),
                ignore_errors=False,
            )

        else:
            raise ValueError(f"Unknown task: {self.task}")

        self.small_mol = True

    def validate_input_molecule(
        self, sequence: str, input_type: str = MoleculeFormat.selfies
    ) -> None:
        """
        Verifies that the non-numerical part of the input sequence is a molecule.

        Args:
            sequence: input sequence to be validated.
            input_type: whether the input is validated to be a SELFIES (default), SMILES or COPOLYMER.
        """

        if input_type == MoleculeFormat.selfies:
            # Fractional molecules based on non-masked parts of the SELFIES sequence
            smis = list(map(decoder, sequence.split(self.tokenizer.mask_token)))
            _, idxs = validate_molecules(smis, input_type)  # type: ignore
        elif (
            input_type == MoleculeFormat.smiles
            or input_type == MoleculeFormat.copolymer
        ):
            _, idxs = validate_molecules([sequence], input_type)
            if len(idxs) != 1:
                raise ValueError(
                    f"The context {sequence} is not a valid {input_type} string."
                )
        else:
            raise ValueError(f"Unknown data type {input_type}.")

    def validate_output(self, sequences: List[Any]) -> Tuple[List[Any], List[int]]:
        """
        Validate the output of the RT model.

        Args:
            sequences: list of sequences to be validated.

        Returns:
            A tuple of validated items:
                - the validate items, a list of either:
                    - Chem.rdchem.Mol (generation task)
                    - Sequence denoting the predicted properties (regression task)
                - list of valid indexes.
        """

        if self.task == "regression":
            return self.validate_numerical(sequences)
        else:
            # Convert SELFIES to SMILES
            if isinstance(self.tokenizer.text_tokenizer, PolymerGraphTokenizer):
                # Copolymer models require specific validation
                return validate_molecules(
                    pattern_list=list(zip(*sequences))[0],
                    input_type=MoleculeFormat.copolymer,
                )

            smiles_list = list(
                filter(
                    lambda x: x != self.target and x is not None,
                    list(zip(*sequences))[0],
                )
            )
            if smiles_list == []:
                return ([None], [-1])
            return validate_molecules(pattern_list=smiles_list, input_type=MoleculeFormat.smiles)  # type: ignore

    def get_maskable_tokens(self, tokens_to_mask: List[str]) -> List[str]:
        """
        Convert a user-defined list of maskable tokens into a RT model-friendly format.

        Args:
            tokens_to_mask: List of atoms specified in SMILES notation.

        Returns:
            List of atoms in SELFIES notation.
        """
        # To reflect double bonds
        tokens_to_mask.extend([f"={t}" for t in tokens_to_mask])
        return [encoder(a) for a in tokens_to_mask]  # type: ignore

    def language_encoding(self, seq: str) -> str:
        if isinstance(self.tokenizer.text_tokenizer, PolymerGraphTokenizer):
            return seq
        selfie = encoder(seq)
        if not isinstance(selfie, str):
            raise TypeError(f"{seq} (type={type(seq)}) is not a valid SMILES sequence.")
        return selfie

    def filter_substructures(
        self, property_successes: Tuple[Tuple[str, str]]
    ) -> Tuple[Tuple[str, str]]:
        """
        Remove samples where user-required substructures are absent from generated samples.

        Args:
            property_successes: A tuple of samples that passed the property constraints.
                This is a tuple of tuples, where the first element is the SMILES and the
                second is the property prediction string.

        Returns:
            A tuple of samples that passed the property constraints and the substructure constraints.
            Same format as input
        """
        if isinstance(self.tokenizer.text_tokenizer, PolymerGraphTokenizer):
            return property_successes
        if self.sampling_wrapper == {}:
            return property_successes

        # Remove samples with stub-like sequences (spuriously small)
        property_successes = filter_stubbed(
            property_successes, target=self.target, threshold=0.5
        )

        successes: List[Tuple[str, str]] = []

        subs_mols = []
        for keep in self.substructures_to_keep:
            subs_mols.append(Chem.MolFromSmiles(keep) or Chem.MolFromSmarts(keep))
            if subs_mols[-1] is None:
                logger.warning(
                    f"{keep} is not a valid SMILES/SELFIES. Instead substructure filtering "
                    f"based on sequence alone can be done and is set to: {self.text_filtering}"
                )

        # Perform filtering
        for smi, prop in property_successes:
            if smi is None:
                continue
            sane = True
            mol = Chem.MolFromSmiles(smi)
            for subs_mol, subs_string in zip(subs_mols, self.substructures_to_keep):
                if subs_mol is None:
                    if self.text_filtering and subs_string not in smi:
                        sane = False
                        break
                else:
                    if not mol.HasSubstructMatch(subs_mol):
                        # Desired substructure not found
                        sane = False
                        break
            if sane:
                successes.append((smi, prop))

        result = cast(Tuple[Tuple[str, str]], tuple(successes))
        return result


class ProteinLanguageRT(ConditionalGenerator):
    """
    Hybrid regression and conditional protein generation model as implemented in
    https://arxiv.org/abs/2202.01338. It generates peptides with a desired stability
    score or predicts the stability score of a given molecule.
    For details on the stability task see: https://doi.org/10.1126/science.aan0693

    Attributes:
        resources_path: path to the model.
        context: user-specified input text for the model.
        search: search key to instantiate a search via terminator.search.SEARCH_FACTORY.
        temperature: the temperature parameter in case of a `sample` search.
        batch_size: the batch size for the model, applicable only to generative task.
        tolerance: the tolerance for the property of the generated molecules.
    """

    subs_format: str = "AAS"

    def __init__(
        self,
        resources_path: str,
        context: str,
        search: str = "sample",
        temperature: float = 1.4,
        batch_size: int = 32,
        tolerance: Union[float, Dict[str, float]] = 20.0,
        sampling_wrapper: Dict[str, Any] = {},
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """
        Initialize the protein generator.

        Args:
            resources_path: directory where to find models and parameters.
            search: search key to instantiate a search, defaults to `sample`.
            temperature: temperature for the sampling. Defaults to 1.4.
            batch_size: number of points sampled per call. Defaults to 8.
            tolerance: the tolerance for the property of the generated molecules.
                Given in percent. Defaults to 20.0.
            sampling_wrapper: A high-level entry point that allows specifying a seed
                SMILES alongside some target conditions.
                NOTE: If this is used, the `target` needs to be a single SMILES string.
                Example: {
                    'fraction_to_mask': 0.5,
                    'tokens_to_mask': [],
                    'property_goal': {'<stab>': 0.85}
                }
            device: device where the inference s running either as a dedicated class
                or a string. If not provided is inferred.
        """
        super().__init__(
            device=device, resources_path=resources_path, tolerance=tolerance
        )

        if sampling_wrapper == {}:
            # Validate input and determine task
            self.task = self.safely_determine_task(context)
        else:
            self.task = "generation"
            self.validate_sampling_wrapper(context=context, **sampling_wrapper)
        self.sampling_wrapper = sampling_wrapper

        # Console outputs for usage of search methods
        if search == "sample" and self.task == "regression":
            logger.warning("For regression task, greedy search is recommended")
        elif search == "greedy" and self.task == "generation":
            logger.warning("For generation task, sample search is recommended")
        if search not in SEARCH_FACTORY.keys():
            raise KeyError(f"Pick a search of {SEARCH_FACTORY.keys()} not: {search}.")
        self.search = SEARCH_FACTORY[search](temperature=temperature)

        # Based on task, set the correct generate_batch method
        if self.task == "regression":
            self.generate_batch = self.generate_batch_regression
        elif self.task == "generation":
            self.generate_batch = self.generate_batch_generation  # type: ignore
            self.batch_size = batch_size
            self.property_collator = PropertyCollator(
                tokenizer=self.tokenizer,
                property_tokens=self.properties,
                num_tokens_to_mask=[-1] * len(self.properties),
                ignore_errors=False,
            )
        else:
            raise ValueError(f"Unknown task: {self.task}")

        self.small_mol = False

    def validate_input_molecule(self, sequence: str, input_type: str = "") -> None:
        """
        Verifies that the non-numerical part of the input sequence is a valid AAS.

        Args:
            sequence: input sequence to be validated.
            input_type: str argument that is ignored but needed for sibling class.
        """
        if sequence != sequence.upper():
            raise ValueError(
                f"Sequence {sequence} does not follow IUPAC convention for AAS"
            )

    def validate_output(self, sequences: List[Any]) -> Tuple[List[Any], List[int]]:
        """
        Validate the output of the RT model.

        Args:
            sequences: list of sequences to be validated.

        Returns:
            A tuple of validated items:
                - List of validated items, either:
                    - Amino acid sequences (generation task)
                    - Sequence denoting the predicted properties (regression task)
                - a list of valid indexes.
        """

        if self.task == "regression":
            return self.validate_numerical(sequences)
        else:
            items = [
                item
                if (
                    (
                        item[0] == item[0].upper()
                        and self.tokenizer.mask_token not in item[0]
                        and not any([s.isdigit() for s in item[0]])
                        and item[0] != self.target
                    )
                    and self.validate_numerical(item[1])
                )
                else None
                for item in sequences
            ]
            idxs = [i for i, item in enumerate(sequences) if item in items]
            return items, idxs

    def get_maskable_tokens(self, tokens_to_mask: List[str]) -> List[str]:
        """
        Convert a user-defined list of maskable tokens (amino acids) into a RT
            model-friendly format. Nothing has to be done for proteins, but the function
            is more complex for sister-classes.

        Args:
            tokens_to_mask: List of amino acids specified in IUPAC convention.

        Returns:
            The same.
        """
        return tokens_to_mask

    def language_encoding(self, seq: str) -> str:
        return seq

    def filter_substructures(
        self, property_successes: Tuple[Tuple[str, str]]
    ) -> Tuple[Tuple[str, str]]:
        """
        Remove samples where user-required substructures are absent from generated samples.

        Args:
            property_successes: A tuple of samples that passed the property constraints.
                This is a tuple of tuples, where the first element is the sequence and the second the
                property prediction string.

        Returns:
            A tuple of samples that passed the property constraints and the substructure constraints.
            Same format as input
        """
        if self.sampling_wrapper == {}:
            return property_successes

        successes = []
        for seq, prop in property_successes:
            sane = True
            for keep in self.substructures_to_keep:
                if keep not in seq:
                    sane = False
                    break
            if sane:
                successes.append((seq, prop))

        return tuple(successes)  # type: ignore



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Data processing utilities."""

import inspect
from typing import Dict, Iterator, List, Optional, Sequence, Tuple, Union, cast

import numpy as np
import torch
from tape.datasets import pad_sequences
from tape.registry import registry
from tape.tokenizers import TAPETokenizer
from torch import nn


class PrimarySequenceEncoder(nn.Module):
    """Model like class to create tape embeddings/encodings.

    This follows tapes implementation via `run_embed` closely, but removes
    any seed/device/cuda handling (of model and batch). This can be done in
    the training loop like for any other nn.Module.

    Example:
        An example use with protein sequence dataset from `pytoda` (requires
        mock/rdkit and pytoda>0.2) passing ids with the primary sequence::

            import sys
            from mock import Mock
            sys.modules['rdkit'] = Mock()
            sys.modules['rdkit.Chem'] = Mock()
            from torch.utils.data import DataLoader
            from pytoda.datasets.protein_sequence_dataset import protein_sequence_dataset
            from pytoda.datasets.tests.test_protein_sequence_dataset import (
                FASTA_CONTENT_GENERIC, TestFileContent
            )
            from pytoda.datasets.utils import keyed

            with TestFileContent(FASTA_CONTENT_GENERIC) as a_test_file:
                sequence_dataset = keyed(protein_sequence_dataset(
                    a_test_file.filename, filetype='.fasta', backend='lazy'
                ))
                batch_size = 5
                dataloader = DataLoader(sequence_dataset, batch_size=batch_size)

                encoder = PrimarySequenceEncoder(
                    model_type='transformer',
                    from_pretrained='bert-base',
                    tokenizer='iupac',
                    log_level=logging.INFO,
                )
                # sending encoder to cuda device should work, not tested

                loaded = next(iter(dataloader))
                print(loaded)
                encoded, ids = encoder.forward(loaded)
                print(ids)
                print(encoded)

        However the forward call supports also not passing ids, but batch still
        has to be wrapped as list (of length 1)::

            encoded, dummy_ids = PrimarySequenceEncoder().forward(
                [
                    ['MQNP', 'LLLLL'],  # type: Sequence[str]
                    # sequence_ids may be missing here
                ]
            )
    """

    def __init__(
        self,
        model_type: str = "transformer",
        from_pretrained: Optional[str] = "bert-base",
        model_config_file: Optional[str] = None,
        # full_sequence_embed: bool = False,
        tokenizer: str = "iupac",
    ):
        """Initialize the PrimarySequenceEncoder.

        Args:
            model_type: Which type of model to create
                (e.g. transformer, unirep, ...). Defaults to 'transformer'.
            from_pretrained: either
                a string with the `shortcut name` of a pre-trained model to
                load from cache or download, e.g.: ``bert-base-uncased``, or
                a path to a `directory` containing model weights saved using
                :func:`tape.models.modeling_utils.ProteinConfig.save_pretrained`,
                e.g.: ``./my_model_directory/``.
                Defaults to 'bert-base'.
            model_config_file: A json config file
                that specifies hyperparameters. Defaults to None.
            tokenizer: vocabulary name. Defaults to 'iupac'.

        Note:
            tapes default seed would be 42 (see `tape.utils.set_random_seeds`)
        """
        super().__init__()
        # padding during forward goes through cpu (numpy)
        self.device_indicator = nn.Parameter(torch.empty(0), requires_grad=False)
        # dummy sequence_ids, so they are optional
        self.next_dummy_id = 0

        task_spec = registry.get_task_spec("embed")  # task = 'embed'
        # from tape.datasets import EmbedDataset
        self.model = registry.get_task_model(
            model_type, task_spec.name, model_config_file, from_pretrained
        )

        # to filter out batch items that aren't used in this model
        # see `from_collated_batch` and `tape.training.ForwardRunner`
        forward_arg_keys = inspect.getfullargspec(self.model.forward).args
        self._forward_arg_keys = forward_arg_keys[1:]  # remove self argument
        assert "input_ids" in self._forward_arg_keys

        self.tokenizer = TAPETokenizer(vocab=tokenizer)
        self.full_sequence_embed = False

        self.eval()

    def train(self, mode: bool):  # type:ignore
        """Avoid any setting to train mode."""
        return super().train(False)

    def generate_tokenized(
        self, batch: List[Sequence[str]]
    ) -> Iterator[Tuple[str, np.ndarray, np.ndarray]]:
        # batch is list of len 2 (typically tuples[str] of length `batch_size`)
        for item, sequence_id in zip(*batch):
            token_ids = self.tokenizer.encode(item)
            input_mask: np.ndarray = np.ones_like(token_ids)
            yield sequence_id, token_ids, input_mask

    @classmethod
    def collate_fn(
        cls, batch: List[Tuple[str, np.ndarray, np.ndarray]]
    ) -> Dict[str, Union[List[str], torch.Tensor]]:
        # from tape.datasets.EmbedDataset because there it's not a classmethod
        ids, tokens, input_mask = zip(*batch)
        ids_list: List[str] = list(ids)
        tokens_tensor: torch.Tensor = torch.from_numpy(pad_sequences(tokens))
        input_mask_tensor: torch.Tensor = torch.from_numpy(pad_sequences(input_mask))
        # on cpu now, is unavoidable as tokenizer and mask are in numpy.
        return {
            "ids": ids_list,
            "input_ids": tokens_tensor,
            "input_mask": input_mask_tensor,
        }  # type: ignore

    def from_collated_batch(
        self, batch: Dict[str, Union[List[str], torch.Tensor]]
    ) -> Dict[str, torch.Tensor]:
        # filter arguments
        batch_tensors: Dict[str, torch.Tensor] = {
            name: tensor  # type:ignore
            for name, tensor in batch.items()
            if name in self._forward_arg_keys
        }
        device = self.device_indicator.device
        if device.type == "cuda":
            batch_tensors = {
                name: tensor.cuda(device=device, non_blocking=True)
                for name, tensor in batch_tensors.items()
            }
        return batch_tensors

    def forward(  # type:ignore
        self, batch: List[Sequence[str]]
    ) -> Tuple[torch.Tensor, List[str]]:
        # batch: List[(primary_sequences,), (sequence_ids,))] of length 2
        # keys can be passed on by pytoda via keyed(ds: Keydataset[str])
        if len(batch) == 1:
            # no sequence_ids passed
            dummy_ids = self.get_dummy_ids(length=len(batch[0]))
            batch.append(dummy_ids)
        elif len(batch) == 2:
            pass
        else:
            raise ValueError(
                "batch should be of length 1 or 2, containing `primary_sequences` "
                " and optionally `sequence_ids`."
            )

        with torch.no_grad():
            # Iterator[(sequence_id, token_ids, input_mask)]
            batch_loader_like = self.generate_tokenized(batch)
            batch_dict_with_ids: Dict[
                str, Union[List[str], torch.Tensor]
            ] = self.collate_fn(list(batch_loader_like))
            ids: List[str] = cast(List[str], batch_dict_with_ids["ids"])
            batch_dict = self.from_collated_batch(batch_dict_with_ids)
            # outputs = self.model(**batch_dict)
            # pooled_embed = outputs[1]
            sequence_embed = self.model(**batch_dict)[0]
            sequence_lengths = batch_dict["input_mask"].sum(1)

            # can variable length slicing be done on the batch?
            if not self.full_sequence_embed:
                sequences_out: torch.Tensor = sequence_embed.new_empty(
                    # dimension of sequence length will be averaged out
                    size=sequence_embed.shape[::2]
                )
            else:
                raise NotImplementedError

            for i, (seqembed, length) in enumerate(
                zip(
                    sequence_embed,
                    sequence_lengths,
                )
            ):
                seqembed = seqembed[: int(length)]
                if not self.full_sequence_embed:
                    seqembed = seqembed.mean(0)
                sequences_out[i, ...] = seqembed

        return sequences_out, ids

    def get_dummy_ids(self, length: int) -> Tuple[str, ...]:
        first = self.next_dummy_id
        self.next_dummy_id += length  # before last
        return tuple(map(str, range(first, self.next_dummy_id)))



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Bases classes and core code used across multiple algorithms."""

from __future__ import annotations

import collections
import logging
import os
import shutil
from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from functools import partial
from time import time
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    Iterator,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
)

from ..configuration import (
    GT4SDConfiguration,
    get_algorithm_subdirectories_in_cache,
    get_algorithm_subdirectories_with_s3,
    get_cached_algorithm_path,
    sync_algorithm_with_s3,
    upload_to_s3,
)
from ..exceptions import GT4SDTimeoutError, InvalidItem, S3SyncError, SamplingError
from ..training_pipelines.core import TrainingPipelineArguments

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

gt4sd_configuration_instance = GT4SDConfiguration.get_instance()

# leave typing generic for algorithm implementation
S = TypeVar("S")  # used for generated items
T = TypeVar("T")  # used for target of generation
U = TypeVar("U")  # used for additional context (e.g. part of target definition)

# callable taking a target
Targeted = Callable[[T], Iterable[Any]]
# callable not taking any target
Untargeted = Callable[[], Iterable[Any]]
# predictive model
Predictor = Callable[[Any], Any]


class GeneratorAlgorithm(ABC, Generic[S, T]):
    """Interface for automated generation via an :class:`AlgorithmConfiguration`."""

    generator: Union[Untargeted, Targeted[T]]
    target: Optional[T]

    #: The maximum amount of time we should let the algorithm run
    max_runtime: int = gt4sd_configuration_instance.gt4sd_max_runtime
    #: The maximum number of samples a user can try to run in one go
    max_samples: int = gt4sd_configuration_instance.gt4sd_max_number_of_samples

    generate: Untargeted

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T] = None,
    ):
        """Targeted or untargeted generation.

        Args:
            configuration: application specific helper that allows to setup the
                generator.
            target: context or condition for the generation. Defaults to None.
        """
        logger.info(
            f"runnning {self.__class__.__name__} with configuration={configuration}"
        )
        generator = self.get_generator(configuration, target)
        setattr(
            self,
            "generate",
            self._setup_untargeted_generator(
                configuration=configuration, generator=generator, target=target
            ),
        )

    @abstractmethod
    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Union[Untargeted, Targeted[T]]:
        """Set up the detail implementation using the configuration.

        Note:
            This is the major method to implement in child classes, it is called
            at instantiation of the GeneratorAlgorithm and must return a callable:

            - Either :obj:`Untargeted`: the callable is taking no arguements,
              and target has to be :obj:`None`.
            - Or :obj:`Targeted`: the callable with the target (but not :obj:`None`).

        Args:
            configuration: application specific helper that allows to setup the
                generator.
            target: context or condition for the generation. Defaults to None.

        Returns:
            generator, the detail implementation used for generation.
            If the target is None, the generator is assumed to be untargeted.
        """

    def timeout(self, item_set: Set, detail: str, error: TimeoutError) -> None:
        """Throws a timeout exception if applicable, otherwise returns
            items gracefully.

        Args:
            item_set: Set of items generated thus far.
            detail: context or condition for the generation.
            error: An error instance, child class of `TimeoutError` either
                `GT4SDTimeoutError` or `SamplingError`.

        Raises:
            TimeoutError: If no items were sampled so far.
        """
        if len(item_set) == 0:
            logger.error(detail + "Exiting now!")
            raise error(title="No samples generated", detail=detail)  # type: ignore
        logger.warning(
            detail
            + f"Returning {len(item_set)} instead of {self.number_of_items} items!"
        )

    def _setup_untargeted_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        generator: Union[Untargeted, Targeted[T]],
        target: Optional[T] = None,
    ) -> Untargeted:
        """Targeted or untargeted generation.

        Args:
            configuration: application specific helper that allows to setup the
                generator.
            generator: the detail implementation used for generation.
                If the target is None, the generator is assumed to be untargeted.
            target: context or condition for the generation. Defaults to None.
        """
        self.configuration = configuration
        self.target = target
        self.generator = generator

        if target is None:
            return self.generator  # type: ignore
        else:
            return partial(self.generator, self.target)  # type: ignore

    def sample(self, number_of_items: int = 100) -> Iterator[S]:
        """Generate a number of unique and valid items.

        Filters duplicate items and iterates batches of generated items to reach
        the desired number of samples, but the number of yielded items is not
        guaranteed:
        In case the generate method does not create new samples for
        GT4SD_MAX_NUMBER_OF_STUCK_CALLS times, it will terminate the
        sampling process.

        Args:
            number_of_items: number of items to generate.
                Defaults to 100.

        Raises:
            SamplingError: when requesting too many items.
            GT4SDTimeoutError: when the algorithm takes longer than the allowed time limit.
                Or when no items were yielded (i.e., if not generating samples for many
                consecutive calls).

        Yields:
            the items.
        """
        self.number_of_items = number_of_items

        if number_of_items > self.max_samples:
            detail = (
                f"{number_of_items} is too many items to generate, "
                f"must be under {self.max_samples+1} samples."
            )
            self.timeout(set(), detail=detail, error=SamplingError)  # type: ignore

        item_set: Set = set()
        stuck_counter = 0
        item_set_length = 0
        start = time()
        while True:
            if time() - start > self.max_runtime:
                detail = f"Sampling took longer than maximal runtime ({self.max_runtime} seconds). "
                self.timeout(item_set, detail=detail, error=GT4SDTimeoutError)  # type: ignore
                return

            generated_items = self.generate()  # type:ignore
            for index, item in enumerate(generated_items):
                try:
                    valid_item = self.configuration.validate_item(item)
                    # check if sample is hashable
                    if not isinstance(item, collections.Hashable):
                        yield valid_item
                        item_set.add(str(index))
                    else:
                        # validation for samples represented as strings
                        if item in item_set:
                            continue
                        else:
                            yield valid_item
                            item_set.add(item)  # type:ignore
                    if len(item_set) == number_of_items:
                        return
                except InvalidItem as error:
                    logger.debug(
                        f"item {item} could not be validated, "
                        f"raising {error.title}: {error.detail}"
                    )
                    continue

            # make sure we don't keep sampling more than a given number of times,
            # in case no new items are generated.
            if len(item_set) == item_set_length:
                stuck_counter += 1
            else:
                stuck_counter = 0
            if (
                stuck_counter
                >= gt4sd_configuration_instance.gt4sd_max_number_of_stuck_calls
            ):
                prefix = "No" if len(item_set) == 0 else "No new"
                detail = f"{prefix} samples generated for more than {gt4sd_configuration_instance.gt4sd_max_number_of_stuck_calls} cycles. "
                self.timeout(item_set, detail=detail, error=GT4SDTimeoutError)  # type: ignore
                return
            item_set_length = len(item_set)

    def validate_configuration(
        self, configuration: AlgorithmConfiguration
    ) -> AlgorithmConfiguration:
        """Overload to validate the a configuration for the algorithm.

        Args:
            configuration: the algorithm configuration.

        Raises:
            InvalidAlgorithmConfiguration: in case the configuration for the algorithm is invalid.

        Returns:
            the validated configuration.
        """
        logger.info("no parameters validation")
        return configuration


class PredictorAlgorithm(ABC, Generic[S, T]):
    """Interface for automated prediction via an :class:`AlgorithmConfiguration`."""

    #: The maximum amount of time we should let the algorithm run
    max_runtime: int = gt4sd_configuration_instance.gt4sd_max_runtime

    def __init__(self, configuration: AlgorithmConfiguration[S, T]):
        """Targeted or untargeted generation.

        Args:
            configuration: application specific helper that allows to setup the
                generator.
        """
        logger.info(
            f"runnning {self.__class__.__name__} with configuration={configuration}"
        )
        self.configuration = configuration
        self.predictor = self.get_predictor(configuration)

    def get_predictor(
        self,
        configuration: AlgorithmConfiguration[S, T],
    ) -> Predictor:
        """Set up the predictive model from the configuration. This is called at
            instantiation of the PredictorAlgorithm and must return a callable.

        Returns:
            predictor, a callable that takes an item and returns a prediction.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        model: Predictor = self.get_model(self.local_artifacts)
        return model

    @abstractmethod
    def get_model(self, resources_path: str) -> Predictor:
        """
        Restore the model from a local path.

        Note:
            This is the major method to implement in child classes, it is called
            at instantiation of the PredictorAlgorithm and must return a callable:

        Args:
            resources_path: local path to the downloaded artifacts.

        Returns:
            Predictor (callable)
        """
        raise NotImplementedError("Not implemented in baseclass.")

    def predict(self, input: Any) -> Any:
        """Perform a prediction for an input.
        Args:
            input: the input for the predictive model

        Raises:
            TimeoutError: when the walltime limit is hit.

        Returns:
            the prediction.
        """

        try:
            predicted = self.predictor(input)
        except TimeoutError:
            detail = (
                f"Predicting took longer than maximum ({self.max_runtime} seconds)."
            )
            logger.warning(detail + " Exiting now!")
        except Exception:
            raise Exception(f"{self.__class__.__name__} failed with {input}")
        return predicted

    def __call__(self, input: Any) -> Any:
        """Alias for `self.predict`."""
        return self.predict(input)


@dataclass
class AlgorithmConfiguration(Generic[S, T]):
    """Algorithm parameter definitions and implementation setup.

    The signature of this class constructor (given by the instance attributes) is used
    for the REST API and needs to be serializable.

    Child classes will add additional instance attributes to configure their respective
    algorithms. This will require setting default values for all of the attributes defined
    here.
    However, the values for :attr:`algorithm_name` and :attr:`algorithm_application`
    are set the registering decorator.

    This strict setup has the following desired effects:

    - Ease child implementation. For example::

        from typing import ClassVar

        from gt4sd.algorithms.registry import ApplicationsRegistry
        from gt4sd.algorithms.core import AlgorithmConfiguration

        @ApplicationsRegistry.register_algorithm_application(ChildOfGeneratorAlgorithm)
        class ConfigurationForChildOfGeneratorAlgorithm(AlgorithmConfiguration):
            algorithm_type: ClassVar[str] = 'generation'
            domain: ClassVar[str] = 'materials'
            algorithm_version: str = 'version3.14'
            actual_parameter: float = 1.61

            # no __init__ definition required


    2. Retrieve the algorithm and configuration easily (via the four class attributes)
    from the :class:`ApplicationsRegistry<gt4sd.algorithms.registry.ApplicationsRegistry>`.
    For example::

       from gt4sd.algorithms.registry import ApplicationsRegistry

       application = ApplicationsRegistry.get_application(
            algorithm_type='generation',
            domain='materials',
            algorithm_name='ChildOfGeneratorAlgorithm',
            algorithm_application='ConfigurationForChildOfGeneratorAlgorithm',
        )
        Algorithm = application.algorithm_class
        Configuration = application.configuration_class

    3. An effortless validation at instantiation via :mod:`pydantic`.

    4. An effortless mapping to artifacts on s3, see :meth:`ensure_artifacts`.

    Todo:
        show how to register a configuration manually (in case it applies to multiple
        algorithms and/or applications)

    """

    #: General type of generative algorithm.
    algorithm_type: ClassVar[str]
    #: General application domain. Hints at input/output types.
    domain: ClassVar[str]
    #: Name of the algorithm to use with this configuration.
    #:
    #: Will be set when registering to :class:`ApplicationsRegistry<gt4sd.algorithms.registry.ApplicationsRegistry>`
    algorithm_name: ClassVar[str]
    #: Unique name for the application that is the use of this
    #: configuration together with a specific algorithm.
    #:
    #: Will be set when registering to :class:`ApplicationsRegistry<gt4sd.algorithms.registry.ApplicationsRegistry>`,
    #: but can be given by direct registration (See :meth:`register_algorithm_application<gt4sd.algorithms.registry.ApplicationsRegistry.register_algorithm_application>`)
    algorithm_application: ClassVar[str]

    #: To differentiate between different versions of an application.
    #:
    #: There is no imposed naming convention.
    algorithm_version: str = ""

    def get_target_description(self) -> Optional[Dict[str, str]]:
        """Get description of the target for generation.

        Returns:
            target description, returns None in case no target is used.
        """
        return {
            "title": "Target for generation",
            "type": "object",
            "description": "Optional target for generation.",
        }

    def to_dict(self) -> Dict[str, Any]:
        """Represent the configuration as a dictionary.

        Returns:
            description of the configuration with parameters description.
        """
        base_configuration_fields_set = set(
            AlgorithmConfiguration.__dataclass_fields__.keys()  # type:ignore
        )
        application_configuration_dict = dict(description=self.__doc__)
        for name, base_description in self.__pydantic_model__.schema()[  # type:ignore
            "properties"
        ].items():
            if name not in base_configuration_fields_set:
                description = dict(
                    getattr(
                        self.__dataclass_fields__[name], "metadata", {}  # type:ignore
                    )
                )
                description.update(base_description)
                if "default" in description:
                    description["optional"] = True
                else:
                    description["optional"] = False
                application_configuration_dict[name] = description  # type:ignore
        return application_configuration_dict

    def validate_item(self, item: Any) -> S:
        """Overload to validate an item.

        Args:
            item: validate an item.

        Raises:
            InvalidItem: in case the item can not be validated.

        Returns:
            S: the validated item.
        """
        # no item validation
        return item

    @classmethod
    def get_application_prefix(cls) -> str:
        """Get prefix up to the specific application.

        Returns:
            the application prefix.
        """
        return os.path.join(
            cls.algorithm_type, cls.algorithm_name, cls.algorithm_application
        )

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        S3 is searched as well as the local cache is searched for matching versions.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """

        prefix = cls.get_application_prefix()
        try:
            versions = get_algorithm_subdirectories_with_s3(prefix)
        except (KeyError, S3SyncError) as error:
            logger.info(
                f"searching S3 raised {error.__class__.__name__}, using local cache only."
            )
            logger.debug(error)
            versions = set()
        versions = versions.union(get_algorithm_subdirectories_in_cache(prefix))
        return versions

    @classmethod
    def list_remote_versions(cls, prefix) -> Set[str]:
        """Get possible algorithm versions on s3.
           Before uploading an artifact on S3, we need to check that
           a particular version is not already present and overwrite by mistake.
           If the final set is empty we can then upload the folder artifact.
           If the final set is not empty, we need to check that the specific version
           of interest is not present.

        only S3 is searched (not the local cache) for matching versions.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        # all name without version
        if not prefix:
            prefix = cls.get_application_prefix()
        try:
            versions = get_algorithm_subdirectories_with_s3(prefix)
        except (KeyError, S3SyncError) as error:
            logger.info(
                f"searching S3 raised {error.__class__.__name__}. This means that no versions are available on S3."
            )
            logger.debug(error)
            versions = set()
        return versions

    @classmethod
    def get_filepath_mappings_for_training_pipeline_arguments(
        cls, training_pipeline_arguments: TrainingPipelineArguments
    ) -> Dict[str, str]:
        """Ger filepath mappings for the given training pipeline arguments.

        Args:
            training_pipeline_arguments: training pipeline arguments.

        Raises:
            ValueError: in case no mapping is available.

        Returns:
            a mapping between artifacts' files and training pipeline's output files.
        """
        raise ValueError(
            f"{cls.__name__} artifacts not mapped for {training_pipeline_arguments}"
        )

    @classmethod
    def save_version_from_training_pipeline_arguments_postprocess(
        cls,
        training_pipeline_arguments: TrainingPipelineArguments,
    ):
        """Postprocess after saving.

        Args:
            training_pipeline_arguments: training pipeline arguments.
        """
        pass

    @classmethod
    def save_version_from_training_pipeline_arguments(
        cls,
        training_pipeline_arguments: TrainingPipelineArguments,
        target_version: str,
        source_version: Optional[str] = None,
    ) -> None:
        """Save a version using training pipeline arguments.

        Args:
            training_pipeline_arguments: training pipeline arguments.
            target_version: target version used to save the model in the cache.
            source_version: source version to use for missing artifacts.
                Defaults to None, a.k.a., use the default version.
        """
        filepaths_mapping: Dict[str, str] = {}
        try:
            filepaths_mapping = (
                cls.get_filepath_mappings_for_training_pipeline_arguments(
                    training_pipeline_arguments=training_pipeline_arguments
                )
            )
        except ValueError:
            logger.info(
                f"{cls.__name__} can not save a version based on {training_pipeline_arguments}"
            )
        if len(filepaths_mapping) > 0:
            if source_version is None:
                source_version = cls.algorithm_version
            source_missing_path = cls.ensure_artifacts_for_version(source_version)
            target_path = os.path.join(
                get_cached_algorithm_path(),
                cls.get_application_prefix(),
                target_version,
            )
            filepaths_mapping = {
                filename: source_filepath
                if os.path.exists(source_filepath)
                else os.path.join(source_missing_path, filename)
                for filename, source_filepath in filepaths_mapping.items()
            }
            logger.info(f"Saving artifacts into {target_path}...")
            try:
                os.makedirs(target_path)
            except OSError:
                logger.warning(
                    f"Artifacts already existing in {target_path}, overwriting them..."
                )
                os.makedirs(target_path, exist_ok=True)
            for target_filename, source_filepath in filepaths_mapping.items():
                target_filepath = os.path.join(target_path, target_filename)
                logger.info(
                    f"Saving artifact {source_filepath} into {target_filepath}..."
                )
                shutil.copyfile(source_filepath, target_filepath)

            cls.save_version_from_training_pipeline_arguments_postprocess(
                training_pipeline_arguments
            )

            logger.info(f"Artifacts saving completed into {target_path}")

    @classmethod
    def upload_version_from_training_pipeline_arguments_postprocess(
        cls,
        training_pipeline_arguments: TrainingPipelineArguments,
    ):
        """Postprocess after uploading. Not implemented yet.

        Args:
            training_pipeline_arguments: training pipeline arguments.
        """
        pass

    @classmethod
    def upload_version_from_training_pipeline_arguments(
        cls,
        training_pipeline_arguments: TrainingPipelineArguments,
        target_version: str,
        source_version: Optional[str] = None,
    ) -> None:
        """Upload a version using training pipeline arguments.

        Args:
            training_pipeline_arguments: training pipeline arguments.
            target_version: target version used to save the model in s3.
            source_version: source version to use for missing artifacts.
                Defaults to None, a.k.a., use the default version.
        """
        filepaths_mapping: Dict[str, str] = {}

        try:
            filepaths_mapping = (
                cls.get_filepath_mappings_for_training_pipeline_arguments(
                    training_pipeline_arguments=training_pipeline_arguments
                )
            )
        except ValueError:
            logger.info(
                f"{cls.__name__} can not save a version based on {training_pipeline_arguments}"
            )

        if len(filepaths_mapping) > 0:
            # probably redundant
            if source_version is None:
                source_version = cls.algorithm_version
            source_missing_path = cls.ensure_artifacts_for_version(source_version)

            # prefix for a run
            prefix = cls.get_application_prefix()
            # versions in s3 with that prefix
            versions = cls.list_remote_versions(prefix)

            # check if the target version is already in s3. If yes, don't upload.
            if target_version not in versions:
                logger.info(
                    f"There is no version {target_version} in S3, starting upload..."
                )
            else:
                logger.info(
                    f"Version {target_version} already exists in S3, skipping upload..."
                )
                return

            # mapping between filenames and paths for a version.
            filepaths_mapping = {
                filename: source_filepath
                if os.path.exists(source_filepath)
                else os.path.join(source_missing_path, filename)
                for filename, source_filepath in filepaths_mapping.items()
            }

            logger.info(
                f"Uploading artifacts into {os.path.join(prefix, target_version)}..."
            )
            try:
                for target_filename, source_filepath in filepaths_mapping.items():
                    # algorithm_type/algorithm_name/algorithm_application/version/filename
                    # for the moment we assume that the prefix exists in s3.
                    target_filepath = os.path.join(
                        prefix, target_version, target_filename
                    )
                    upload_to_s3(target_filepath, source_filepath)
                    logger.info(
                        f"Upload artifact {source_filepath} into {target_filepath}..."
                    )

            except S3SyncError:
                logger.warning("Problem with upload...")
                return

            logger.info(
                f"Artifacts uploading completed into {os.path.join(prefix, target_version)}"
            )

    @classmethod
    def ensure_artifacts_for_version(cls, algorithm_version: str) -> str:
        """The artifacts matching the path defined by class attributes and the given version are downloaded.

        That is all objects under ``algorithm_type/algorithm_name/algorithm_application/algorithm_version``
        in the bucket are downloaded.

        Args:
            algorithm_version: version of the algorithm to ensure artifacts for.

        Returns:
            the common local path of the matching artifacts.
        """
        prefix = os.path.join(
            cls.get_application_prefix(),
            algorithm_version,
        )
        try:
            local_path = sync_algorithm_with_s3(prefix)
        except (KeyError, S3SyncError) as error:
            logger.info(
                f"searching S3 raised {error.__class__.__name__}, using local cache only."
            )
            logger.debug(error)
            local_path = get_cached_algorithm_path(prefix)
            if not os.path.isdir(local_path):
                raise OSError(
                    f"artifacts directory {local_path} does not exist locally, and syncing with s3 failed: {error}"
                )

        return local_path

    def ensure_artifacts(self) -> str:
        """The artifacts matching the path defined by class attributes are downloaded.

        That is all objects under ``algorithm_type/algorithm_name/algorithm_application/algorithm_version``
        in the bucket are downloaded.

        Returns:
            the common local path of the matching artifacts.
        """
        return self.ensure_artifacts_for_version(self.algorithm_version)


class ConfigurablePropertyAlgorithmConfiguration(AlgorithmConfiguration):
    """A configurable AlgorithmConfiguration to be used by the properties submodule."""

    module: str = "properties"

    def __init__(
        self,
        domain: str,
        algorithm_name: str,
        algorithm_application: str,
        algorithm_version: str = "v0",
        algorithm_type: str = "prediction",
    ):
        """
        Args:
            domain: submodule of properties where the model resides.
            algorithm_version: name of the predictive model, e.g., `MCA`.
            algorithm_application: application of the algorithm, e.g., dataset it was
                trained on, like `Tox21`.
            algorithm_version: version of the algorithm, defaults to `v0`.
            algorithm_type: type of the algorithm. This should be `prediction`.
        """
        self.algorithm_type = algorithm_type  # type: ignore
        self.domain = domain  # type: ignore
        self.algorithm_name = algorithm_name  # type: ignore
        self.algorithm_application = algorithm_application  # type: ignore
        self.algorithm_version = algorithm_version

    def get_application_prefix(self) -> str:  # type: ignore
        """Get prefix up to the specific application.

        NOTE: Unlike the parent method this uses the assgined attributes since it's
        configurable.

        Returns:
            the application prefix.
        """
        return os.path.join(
            self.domain, self.algorithm_name, self.algorithm_application
        )

    def ensure_artifacts_for_version(self, algorithm_version: str) -> str:  # type: ignore
        """The artifacts matching the path defined by class attributes and the given version are downloaded.

        NOTE: Unlike the parent method this uses the assigned attributes since it's
        configurable.

        That is all objects under ``algorithm_type/algorithm_name/algorithm_application/algorithm_version``
        in the bucket are downloaded.

        Args:
            algorithm_version: version of the algorithm to ensure artifacts for.

        Returns:
            the common local path of the matching artifacts.
        """
        prefix = os.path.join(
            self.get_application_prefix(),
            algorithm_version,
        )
        try:
            local_path = sync_algorithm_with_s3(prefix, module=self.module)
        except (KeyError, S3SyncError) as error:
            logger.info(
                f"searching S3 raised {error.__class__.__name__}, using local cache only."
            )
            logger.debug(error)
            local_path = get_cached_algorithm_path(prefix, module=self.module)
            if not os.path.isdir(local_path):
                raise OSError(
                    f"artifacts directory {local_path} does not exist locally, and syncing with s3 failed: {error}"
                )

        return local_path

    def list_versions(self) -> Set[str]:  # type: ignore
        """Get possible algorithm versions.

        NOTE: Unlike the parent method this uses the assigned attributes since it's
        configurable.

        S3 is searched as well as the local cache is searched for matching versions.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """

        prefix = self.get_application_prefix()
        try:
            versions = get_algorithm_subdirectories_with_s3(prefix, module=self.module)
        except (KeyError, S3SyncError) as error:
            logger.info(
                f"searching S3 raised {error.__class__.__name__}, using local cache only."
            )
            logger.debug(error)
            versions = set()
        versions = versions.union(
            get_algorithm_subdirectories_in_cache(prefix, module=self.module)
        )
        return versions


def get_configuration_class_with_attributes(
    klass: Type[AlgorithmConfiguration],
) -> Type[AlgorithmConfiguration]:
    """Get AlgorithmConfiguration with set attributes.

    Args:
        klass: a class to be used to extract attributes from.

    Returns:
        a class with the attributes set.
    """
    configuration_class = deepcopy(AlgorithmConfiguration)
    setattr(configuration_class, "algorithm_type", klass.algorithm_type)
    setattr(configuration_class, "algorithm_name", klass.algorithm_name)
    setattr(configuration_class, "algorithm_application", klass.__name__)
    setattr(configuration_class, "algorithm_version", klass.algorithm_version)
    return configuration_class


class PropertyPredictor(ABC, Generic[S, U]):
    """TODO: Might be deprecated in future release."""

    def __init__(self, context: U) -> None:
        """Property predictor to investigate items.

        Args:
            context: the context in which a property of an item can be
                computed or checked is very application specific.
        """
        logger.warning("This class will be deprecated in a future release.")
        self.context = context
        # or pass these with methods?

    @abstractmethod
    def satisfies(self, item: S) -> bool:
        """Check whether an item satisfies given requirements.

        Args:
            item: the item to check.

        Returns:
            bool:
        """

    def compute(self, item: S) -> Any:
        """Compute some metric/property on an item.

        Args:
            item: the item to compute a metric on.

        Returns:
            Any: the computed metric/property.
        """
